#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
API client
"""
import json
import logging
import random
import string
import sys
import threading
from collections.abc import AsyncIterable
from typing import Mapping, Any, Optional, Iterator, Tuple, Union, Dict, Callable
from urllib.parse import urlparse, urlencode

import aiohttp
import requests
import requests.exceptions

import odahuflow.sdk.config
from odahuflow.sdk.clients.oauth_handler import start_oauth2_callback_handler, OAuthLoginResult, do_refresh_token, \
    do_client_cred_authentication
from odahuflow.sdk.config import update_config_file
from odahuflow.sdk.definitions import API_VERSION

LOGGER = logging.getLogger(__name__)


class WrongHttpStatusCode(Exception):
    """
    Exception for wrong HTTP status code
    """

    def __init__(self, status_code: int, http_result: Dict[str, str] = None):
        """
        Initialize Wrong Http Status Code exception

        :param status_code: HTTP status code
        :param http_result: HTTP data
        """
        if http_result is None:
            http_result = {}
        super().__init__(f'Got error from server: {http_result.get("message")} (status: {status_code})')

        self.status_code = status_code


class APIConnectionException(Exception):
    """
    Exception that says that client can not reach API server
    """

    pass


class IncorrectAuthorizationToken(APIConnectionException):
    """
    Exception that says that provided API authorization token is incorrect
    """

    pass


class IncorrectClientCredentials(APIConnectionException):
    """
    Exception that says that provided API authorization token is incorrect
    """

    pass


class LoginRequired(APIConnectionException):
    """
    Exception that says that login is required to do calls
    """

    pass


def get_authorization_redirect(web_redirect: str, after_login: Callable) -> str:
    """
    Try to detect, parse and build OAuth2 redirect

    :param web_redirect: returned redirect
    :param after_login: function that have to be called after successful login
    :return: str -- new redirect
    """
    loc = urlparse(web_redirect)

    state = ''.join(random.choice(string.ascii_letters) for _ in range(10))
    local_check_address = start_oauth2_callback_handler(after_login, state, web_redirect)

    get_parameters = {
        'client_id': odahuflow.sdk.config.ODAHUFLOWCTL_OAUTH_CLIENT_ID,
        'response_type': 'code',
        'state': state,
        'redirect_uri': local_check_address,
        'scope': odahuflow.sdk.config.ODAHUFLOWCTL_OAUTH_SCOPE
    }
    web_redirect = f'{loc.scheme}://{loc.netloc}{loc.path}?{urlencode(get_parameters)}'
    return web_redirect


class RemoteAPIClient:
    """
    Base API client
    """

    def __init__(self,
                 base_url: str = odahuflow.sdk.config.API_URL,
                 token: Optional[str] = odahuflow.sdk.config.API_TOKEN,
                 client_id: Optional[str] = '',
                 client_secret: Optional[str] = '',
                 retries: Optional[int] = 3,
                 timeout: Optional[Union[int, Tuple[int, int]]] = 10,
                 non_interactive: Optional[bool] = True):
        """
        Build client

        :param base_url: base url, for example: http://api.example.com
        :param token: token for token based auth
        :param client_id: client_id for Client Credentials OAuth2 flow
        :param client_secret: client_secret for Client Credentials OAuth2 flow
        :param retries: command retries or less then 2 if disabled
        :param timeout: timeout for connection in seconds. 0 for disabling
        :param non_interactive: disable any interaction
        """
        self._base_url = base_url
        self._token = token
        self._client_id = client_id
        self._client_secret = client_secret
        self._version = API_VERSION
        self._retries = retries
        self._non_interactive = non_interactive
        self._interactive_login_finished = threading.Event()

        self.timeout = timeout
        # Force if set
        if odahuflow.sdk.config.ODAHUFLOWCTL_NONINTERACTIVE:
            self._non_interactive = True

    def _update_config_with_new_oauth_config(self, login_result: OAuthLoginResult) -> None:
        """
        Update config with new oauth credentials

        :param login_result: result of login
        :return: None
        """
        self._token = login_result.id_token
        update_config_file(API_URL=self._base_url,
                           API_TOKEN=login_result.id_token,
                           API_REFRESH_TOKEN=login_result.refresh_token,
                           API_ACCESS_TOKEN=login_result.access_token,
                           API_ISSUING_URL=login_result.issuing_url)

    def after_login(self, login_result: OAuthLoginResult) -> None:
        """
        Handle action after login

        :param login_result: result of login
        :return: None
        """
        self._interactive_login_finished.set()
        self._update_config_with_new_oauth_config(login_result)
        print('You has been authorized on endpoint %s as %s / %s' %
                    (self._base_url, login_result.user_name, login_result.user_email))
        sys.exit(0)

    @classmethod
    def construct_from_other(cls, other):
        """
        Construct API-based client from another API-based client

        :param other: API-based client to get connection options from
        :return: self -- new client
        """
        return cls(other._base_url, other._token, other._retries, other.timeout)

    def _build_url(self, url_template):
        sub_url = url_template.format(version=self._version)
        target_url = self._base_url.strip('/') + sub_url
        return target_url

    def _build_request_kwargs(self,
                              url_template: str,
                              payload: Mapping[Any, Any] = None,
                              action: str = 'GET',
                              stream: bool = False) -> Dict[str, Any]:
        target_url = self._build_url(url_template)
        headers = {}
        if self._token:
            headers['Authorization'] = f'Bearer {self._token}'
        if stream:
            headers['Content-type'] = 'text/event-stream'

        request_kwargs = {
            'method': action,
            'url': target_url,
            'params' if action.lower() == 'get' else 'json': payload,
            'headers': headers
        }
        return request_kwargs

    def _build_request(self,
                       url_template: str,
                       payload: Mapping[Any, Any] = None,
                       action: str = 'GET',
                       stream: bool = False) -> requests.PreparedRequest:

        request_kwargs = self._build_request_kwargs(url_template, payload, action, stream)
        request = requests.PreparedRequest()
        request.prepare(**request_kwargs)
        return request

    def _get_conn_timeout(self, timeout=None):
        connection_timeout = timeout if timeout is not None else self.timeout

        if connection_timeout == 0:
            connection_timeout = None

        return connection_timeout

    def _get_left_retries(self):
        return self._retries if self._retries > 0 else 1

    def _request(self,
                 url_template: str,
                 payload: Mapping[Any, Any] = None,
                 action: str = 'GET',
                 stream: bool = False,
                 timeout: Optional[int] = None,
                 limit_stack: bool = False,
                 client: requests.Session = None):
        """
        Make HTTP request

        :param url_template: target URL
        :param payload: data to be placed in body of request
        :param action: request action, e.g. get / post / delete
        :param stream: use stream mode or not
        :param timeout: custom timeout in seconds (overrides default). 0 for disabling
        :param limit_stack: do not start refreshing token if it is possible
        :return: :py:class:`requests.Response` -- response
        """

        request_kwargs = self._build_request_kwargs(url_template, payload, action, stream)
        connection_timeout = self._get_conn_timeout(timeout)
        left_retries = self._get_left_retries()

        raised_exception = None
        while left_retries > 0:
            try:
                LOGGER.debug('Requesting {}'.format(self._build_url(url_template)))

                if client:
                    response = client.request(timeout=connection_timeout, stream=stream, **request_kwargs)
                else:
                    response = requests.request(timeout=connection_timeout, stream=stream, **request_kwargs)

                LOGGER.debug('Response status code: "{}"'.format(response.status_code))

            except requests.exceptions.ConnectionError as exception:
                LOGGER.error('Failed to connect to {}: {}. Retrying'.format(self._base_url, exception))
                raised_exception = exception
            else:
                LOGGER.debug('Got response. Breaking')
                break
            left_retries -= 1
        else:
            raise APIConnectionException('Can not reach {}'.format(self._base_url)) from raised_exception

        if self._login_required(response):
            try:
                self._login(str(response.url), limit_stack=limit_stack)
            except IncorrectAuthorizationToken as login_exc:
                raise login_exc from raised_exception
            return self._request(
                url_template,
                payload=payload,
                action=action,
                stream=stream,
                timeout=timeout,
                limit_stack=True
            )
        else:
            return response

    def query(self, url_template: str, payload: Mapping[Any, Any] = None, action: str = 'GET'):
        """
        Perform query to API server

        :param url_template: url template from odahuflow.const.api
        :param payload: payload (will be converted to JSON) or None
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :return: dict[str, any] -- response content
        """
        response = self._request(url_template, payload, action)
        return self._handle_query_response(response.text, payload, response.status_code)

    @staticmethod
    def _handle_query_response(text: str, payload: Mapping[Any, Any], status_code: int) -> Dict:
        try:
            answer = json.loads(text)
            LOGGER.debug('Got answer: {!r} with code {} for URL {!r}'
                         .format(answer, status_code, payload))
        except ValueError:
            answer = {}

        if 400 <= status_code < 600:
            raise WrongHttpStatusCode(status_code, answer)

        LOGGER.debug('Query has been completed, parsed and validated')
        return answer

    def stream(self, url_template: str, action: str = 'GET', params: Mapping[str, Any] = None) -> Iterator[str]:
        """
        Perform query to API server

        :param url_template: url template from odahuflow.const.api
        :param params: payload (will be converted to JSON) or None
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :return: response content
        """
        response = self._request(url_template, action=action, stream=True, payload=params)

        with response:
            if not response.ok:
                raise WrongHttpStatusCode(response.status_code)

            for line in response.iter_lines():
                yield line.decode("utf-8")

    def info(self):
        """
        Perform info query on API server

        :return:
        """
        try:
            return self.query("/health")
        except ValueError:
            pass

    ############################
    # API login helper methods #
    ############################

    @staticmethod
    def _login_required(response):
        """
        Check whether the login is required or client is already is authorised
        :param response:
        :return:
        """
        # We assume if there were redirects then credentials are out of date and we can refresh or build auth url

        def not_authorized_resp_code():
            if hasattr(response, 'status_code'):
                return response.status_code == 401
            elif hasattr(response, 'status'):
                return response.status == 401
            return False

        return bool(response.history) or not_authorized_resp_code()

    def _login(self, url: str, limit_stack=False):
        """
        Authorise client. Next methods are used by priority:
        1. Refreshing token (if token exists)
        2. Interactive mode (if enabled)
        :return: None if success, IncorrectAuthorizationToken exception otherwise
        """

        # If it is a error after refreshed token - fail
        if limit_stack:
            raise IncorrectAuthorizationToken(
                f'{self._credentials_error_status} even after refreshing. \n'
                'Please try to log in again'
            )

        LOGGER.debug('Redirect has been detected. Trying to refresh a token')
        if self._refresh_token_exists:
            LOGGER.debug('Refresh token for %s has been found, trying to use it', odahuflow.sdk.config.API_ISSUING_URL)
            self._login_with_refresh_token()
        elif self._client_id and self._client_secret and odahuflow.sdk.config.API_ISSUING_URL:
            self._login_with_client_credentials()
        elif self._interactive_mode_enabled:
            # Start interactive flow
            self._login_interactive_mode(url)
        else:
            raise IncorrectAuthorizationToken(
                f'{self._credentials_error_status}. \n'
                'Please provide correct temporary token or disable non interactive mode'
            )

    def _login_with_refresh_token(self):
        login_result = do_refresh_token(odahuflow.sdk.config.API_REFRESH_TOKEN, odahuflow.sdk.config.API_ISSUING_URL)
        if not login_result:
            raise IncorrectAuthorizationToken(
                'Refresh token in not correct. \n'
                'Please login again'
            )
        else:
            self._update_config_with_new_oauth_config(login_result)

    def _login_with_client_credentials(self):
        login_result = do_client_cred_authentication(
            issue_token_url=odahuflow.sdk.config.API_ISSUING_URL, client_id=self._client_id,
            client_secret=self._client_secret
        )
        if not login_result:
            raise IncorrectClientCredentials(
                'Client credentials in not correct. \n'
                'Please login again'
            )
        else:
            self._update_config_with_new_oauth_config(login_result)

    def _login_interactive_mode(self, url):
        self._interactive_login_finished.clear()
        target_url = get_authorization_redirect(url, self.after_login)
        print('%s. \nPlease open %s' % (self._credentials_error_status, target_url))
        self._interactive_login_finished.wait()

    @property
    def _refresh_token_exists(self):
        return odahuflow.sdk.config.API_REFRESH_TOKEN and odahuflow.sdk.config.API_ISSUING_URL

    @property
    def _client_credentials_exists(self):
        return odahuflow.sdk.config.ODAHUFLOWCTL_OAUTH_CLIENT_ID and \
               odahuflow.sdk.config.ODAHUFLOWCTL_OAUTH_CLIENT_SECRET

    @property
    def _interactive_mode_enabled(self):
        return not self._non_interactive

    @property
    def _credentials_error_status(self):
        if self._token:
            credentials_error_status = 'Credentials are not correct'
        else:
            credentials_error_status = 'Credentials are missed'
        return credentials_error_status


class AsyncRemoteAPIClient(RemoteAPIClient):

    async def _async_request(
            self, url_template: str,
            payload: Mapping[Any, Any] = None,
            action: str = 'GET',
            stream=False,
            session: aiohttp.ClientSession = None
    ) -> AsyncIterable:
        """
        Perform async request to API server

        :param url_template: url template from odahuflow.const.api
        :param payload: payload (will be converted to JSON) or None
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :param stream: use stream mode or not
        :param session: aiohttp.Session for making response
        :return: AsyncIterable of data. If stream = False, only one line will be returned
        """
        assert session is not None

        request_kwargs = self._build_request_kwargs(url_template, payload, action, stream=False)
        left_retries = self._get_left_retries()
        raised_exception = None
        while left_retries > 0:
            try:
                async with session.request(**request_kwargs) as resp:
                    print(resp)
                    if self._login_required(resp):
                        raise LoginRequired()

                    if stream:
                        LOGGER.debug('Status code: "{}", Response: "<stream>"'.format(resp.status))
                        async for line in self._handle_stream_response(resp):
                            yield line
                    else:
                        resp_text = await resp.text()
                        LOGGER.debug('Status code: "{}", Response: "{}"'.format(resp.status, resp_text))
                        data = self._handle_query_response(resp_text, payload, resp.status)
                        yield data
                    break
            except aiohttp.ClientConnectionError as exception:
                LOGGER.error('Failed to connect to {}: {}. Retrying'.format(self._base_url, exception))
                raised_exception = exception
                left_retries -= 1
        else:
            raise APIConnectionException('Can not reach {}'.format(self._base_url)) from raised_exception

    async def _handle_stream_response(self, response: aiohttp.ClientResponse, chunk_size=500) -> AsyncIterable:
        """
        Helper method that do next things:
        1) Async iterating over response content by chunks
        2) Decode bytes to text
        3) Build string lines from text stream
        :param response:
        :param chunk_size:
        :return:
        """

        if 400 <= response.status < 600:
            raise WrongHttpStatusCode(response.status)

        encoding = response.get_encoding()
        pending = None

        async for bytes_chunk in response.content.iter_chunked(chunk_size):
            chunk = bytes_chunk.decode(encoding)

            if pending is not None:

                chunk = pending + chunk

            lines = chunk.splitlines()

            if lines and lines[-1] and chunk and lines[-1][-1] == chunk[-1]:
                pending = lines.pop()
            else:
                pending = None

            for line in lines:
                yield line

        if pending is not None:
            yield pending

    async def _async_request_with_login(
            self, url_template: str, payload: Mapping[Any, Any] = None, action: str = 'GET', stream=False
    ):
        """
        Perform async request to API server.
        Create session under the hood.
        Make two attempts to get data. Second attempt is used after login if required

        :param url_template: url template from odahuflow.const.api
        :param payload: payload (will be converted to JSON) or None
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :param stream: use stream mode or not
        :return: AsyncIterable of data. If stream = False, only one line will be returned
        """
        async with aiohttp.ClientSession(conn_timeout=self._get_conn_timeout(), trust_env=True) as session:
            try:
                async for data in self._async_request(url_template, payload, action, stream, session):
                    yield data
            except LoginRequired:
                self._login(self._build_url(url_template))
                async for data in self._async_request(url_template, payload, action, stream, session):
                    yield data

    async def query(self, url_template: str, payload: Mapping[Any, Any] = None, action: str = 'GET') -> Any:
        """
        Perform query to API server.

        :param url_template: url template from odahuflow.const.api
        :param payload: payload (will be converted to JSON) or None
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :return: dict[str, any] -- response content
        """
        resp = None
        async for res in self._async_request_with_login(url_template, payload, action, stream=False):
            resp = res
        return resp

    async def stream(self, url_template: str, action: str = 'GET', params: Mapping[str, Any] = None) -> AsyncIterable:
        """
        Perform query to API server in stream mode

        :param url_template: url template from odahuflow.const.api
        :param action: HTTP method (GET, POST, PUT, DELETE)
        :param params: payload (will be converted to JSON) or None
        :return: async iterable that return response content line by line
        """
        async for line in self._async_request_with_login(url_template, action=action, stream=True, payload=params):
            yield line

    async def info(self):
        """
        Perform info query on API server
        :return:
        """
        try:
            return await self.query("/health")
        except ValueError:
            pass



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class HTTPResult(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, message: str=None):  # noqa: E501
        """HTTPResult - a model defined in Swagger

        :param message: The message of this HTTPResult.  # noqa: E501
        :type message: str
        """
        self.swagger_types = {
            'message': str
        }

        self.attribute_map = {
            'message': 'message'
        }

        self._message = message

    @classmethod
    def from_dict(cls, dikt) -> 'HTTPResult':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The HTTPResult of this HTTPResult.  # noqa: E501
        :rtype: HTTPResult
        """
        return util.deserialize_model(dikt, cls)

    @property
    def message(self) -> str:
        """Gets the message of this HTTPResult.

        Success of error message  # noqa: E501

        :return: The message of this HTTPResult.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message: str):
        """Sets the message of this HTTPResult.

        Success of error message  # noqa: E501

        :param message: The message of this HTTPResult.
        :type message: str
        """

        self._message = message



import datetime
import json
import logging
import os
from os import listdir
from os.path import join
from pathlib import Path
from shutil import rmtree
from typing import List

import docker
from docker.models.containers import Container
from docker.types import Mount
from odahuflow.sdk import config
from odahuflow.sdk.local.docker_utils import stream_container_logs, TRAINING_DOCKER_LABELS, WORKSPACE_PATH, \
    convert_labels_to_filter, cleanup_docker_containers, raise_error_if_container_failed
from odahuflow.sdk.models import K8sTrainer

MODEL_OUTPUT_CONTAINER_PATH = '/output'
TRAINER_CONF_PATH = 'mt.json'

LOGGER = logging.getLogger(__name__)


def launch_training_container(trainer: K8sTrainer, output_dir: str) -> None:
    client = docker.from_env()
    container: Container = client.containers.run(
        trainer.model_training.spec.image or trainer.toolchain_integration.spec.default_image,
        stderr=True,
        stdout=True,
        working_dir=WORKSPACE_PATH,
        command=[
            trainer.toolchain_integration.spec.entrypoint,
            "--verbose",
            "--mt",
            join(WORKSPACE_PATH, TRAINER_CONF_PATH),
            "--target",
            MODEL_OUTPUT_CONTAINER_PATH,
        ],
        mounts=[
            Mount(WORKSPACE_PATH, os.getcwd(), type="bind"),
            Mount(MODEL_OUTPUT_CONTAINER_PATH, output_dir, type="bind")
        ],
        detach=True,
        labels=TRAINING_DOCKER_LABELS,
    )

    container_info = client.api.inspect_container(container.id)
    LOGGER.debug(f'Container info:\n{json.dumps(container_info, indent=2)}')

    print(f'Training docker image {container.id} has started. Stream logs:')
    stream_container_logs(container)

    raise_error_if_container_failed(container.id)


def launch_gppi_validation_container(trainer: K8sTrainer, output_dir: str) -> None:
    client = docker.from_env()
    container: Container = client.containers.run(
        trainer.model_training.spec.image or trainer.toolchain_integration.spec.default_image,
        stderr=True,
        stdout=True,
        working_dir=WORKSPACE_PATH,
        command=[
            "odahuflowctl",
            "--verbose",
            "gppi",
            "--gppi-model-path",
            MODEL_OUTPUT_CONTAINER_PATH,
            "--env-name",
            "odahu_model",
            "test"
        ],
        mounts=[
            Mount(WORKSPACE_PATH, os.getcwd(), type="bind"),
            Mount(MODEL_OUTPUT_CONTAINER_PATH, output_dir, type="bind")
        ],
        detach=True,
        labels=TRAINING_DOCKER_LABELS,
    )

    container_info = client.api.inspect_container(container.id)
    LOGGER.debug(f'Container info:\n{json.dumps(container_info, indent=2)}')

    print(f'Validation docker image {container.id} has started. Stream logs:')
    stream_container_logs(container)

    raise_error_if_container_failed(container.id)


def create_mt_config_file(trainer: K8sTrainer) -> None:
    with open(TRAINER_CONF_PATH, 'w') as f:
        trainer_dict = trainer.to_dict()
        json.dump(trainer_dict, f)

        LOGGER.debug(f"Saved the trainer configuration:\n{json.dumps(trainer_dict, indent=2)}")


def start_train(trainer: K8sTrainer, output_dir: str):
    create_mt_config_file(trainer)

    if not output_dir:
        # For example, 01-Mar-2020-17-38-04
        suffix = datetime.datetime.now().strftime("%d-%b-%Y-%H-%M-%S")
        output_dir = join(
            config.LOCAL_MODEL_OUTPUT_DIR,
            f'{trainer.model_training.spec.model.name}-{trainer.model_training.spec.model.version}-{suffix}'
        )

        LOGGER.debug(f'Output model training directory is not provided. Generate the directory: {output_dir}')

    os.makedirs(output_dir, exist_ok=True)

    launch_training_container(trainer, output_dir)

    launch_gppi_validation_container(trainer, output_dir)

    print(f'Model {Path(output_dir).name} was saved in the {output_dir} directory')


def list_local_trainings() -> List[str]:
    if not os.path.exists(config.LOCAL_MODEL_OUTPUT_DIR):
        return []

    return listdir(config.LOCAL_MODEL_OUTPUT_DIR)


def cleanup_local_artifacts():
    if os.path.exists(config.LOCAL_MODEL_OUTPUT_DIR):
        rmtree(config.LOCAL_MODEL_OUTPUT_DIR)

        LOGGER.debug(f'Remove the {config.LOCAL_MODEL_OUTPUT_DIR} directory')


def cleanup_training_docker_containers():
    cleanup_docker_containers(convert_labels_to_filter(TRAINING_DOCKER_LABELS))


if __name__ == '__main__':
    cleanup_training_docker_containers()



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
odahuflow k8s definitions functions
"""

API_VERSION = 'v1'
CONFIGURATION_URL = '/api/{version}/configuration'
CONNECTION_URL = '/api/{version}/connection'
MODEL_TRAINING_URL = '/api/{version}/model/training'
TOOLCHAIN_INTEGRATION_URL = '/api/{version}/toolchain/integration'
MODEL_DEPLOYMENT_URL = '/api/{version}/model/deployment'
MODEL_ROUTE_URL = '/api/{version}/model/route'
MODEL_PACKING_URL = '/api/{version}/model/packaging'
PACKING_INTEGRATION_URL = '/api/{version}/packaging/integration'



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
API client
"""
from collections.abc import AsyncIterable
from typing import List, Iterator

from odahuflow.sdk.clients.api import RemoteAPIClient, AsyncRemoteAPIClient
from odahuflow.sdk.definitions import MODEL_PACKING_URL
from odahuflow.sdk.models import ModelPackaging

SCHEDULING_STATE = "scheduling"
RUNNING_STATE = "running"
SUCCEEDED_STATE = "succeeded"
FAILED_STATE = "failed"
UNKNOWN_STATE = "unknown"


class ModelPackagingClient(RemoteAPIClient):
    """
    Model Packaging client
    """

    def get(self, name: str) -> ModelPackaging:
        """
        Get Model Packaging from API server

        :param name: Model Packaging name
        :type name: str
        :return: Model Packaging
        """
        return ModelPackaging.from_dict(self.query(f'{MODEL_PACKING_URL}/{name}'))

    def get_all(self) -> List[ModelPackaging]:
        """
        Get all Model Packagings from API server

        :return: all Model Packagings
        """
        return [ModelPackaging.from_dict(mr) for mr in self.query(MODEL_PACKING_URL)]

    def create(self, mr: ModelPackaging) -> ModelPackaging:
        """
        Create Model Packaging

        :param mr: Model Packaging
        :return Message from API server
        """
        return ModelPackaging.from_dict(
            self.query(MODEL_PACKING_URL, action='POST', payload=mr.to_dict())
        )

    def edit(self, mr: ModelPackaging) -> ModelPackaging:
        """
        Edit Model Packaging

        :param mr: Model Packaging
        :return Message from API server
        """
        return ModelPackaging.from_dict(
            self.query(MODEL_PACKING_URL, action='PUT', payload=mr.to_dict())
        )

    def delete(self, name: str) -> str:
        """
        Delete Model Packagings

        :param name: Name of a Model Packaging
        :return Message from API server
        """
        return self.query(f'{MODEL_PACKING_URL}/{name}', action='DELETE')

    def log(self, name: str, follow: bool = False) -> Iterator[str]:
        """
        Stream logs from packaging

        :param follow: follow stream
        :param name: Name of a Model Packaging
        :return Message from API server
        """
        return self.stream(f'{MODEL_PACKING_URL}/{name}/log', 'GET', params={'follow': follow})


class AsyncModelPackagingClient(AsyncRemoteAPIClient, ModelPackagingClient):
    """
    Model Packaging async client
    """

    async def get(self, name: str) -> ModelPackaging:
        """
        Get Model Packaging from API server

        :param name: Model Packaging name
        :type name: str
        :return: Model Packaging
        """
        return ModelPackaging.from_dict(await self.query(f'{MODEL_PACKING_URL}/{name}'))

    async def get_all(self) -> List[ModelPackaging]:
        """
        Get all Model Packagings from API server

        :return: all Model Packagings
        """
        return [ModelPackaging.from_dict(mr) for mr in await self.query(MODEL_PACKING_URL)]

    async def create(self, mr: ModelPackaging) -> ModelPackaging:
        """
        Create Model Packaging

        :param mr: Model Packaging
        :return Message from API server
        """
        return ModelPackaging.from_dict(
            await self.query(MODEL_PACKING_URL, action='POST', payload=mr.to_dict())
        )

    async def edit(self, mr: ModelPackaging) -> ModelPackaging:
        """
        Edit Model Packaging

        :param mr: Model Packaging
        :return Message from API server
        """
        return ModelPackaging.from_dict(
            await self.query(MODEL_PACKING_URL, action='PUT', payload=mr.to_dict())
        )

    async def delete(self, name: str) -> str:
        """
        Delete Model Packagings

        :param name: Name of a Model Packaging
        :return Message from API server
        """
        return await self.query(f'{MODEL_PACKING_URL}/{name}', action='DELETE')

    async def log(self, name: str, follow: bool = False) -> AsyncIterable:
        """
            Stream logs from packaging

            :param follow: follow stream
            :param name: Name of a Model Packaging
            :return Message from API server
            """
        params = {'follow': 'true' if follow else 'false'}
        async for chunk in self.stream(f'{MODEL_PACKING_URL}/{name}/log', 'GET', params=params):
            yield chunk



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models.data_binding_dir import DataBindingDir  # noqa: F401,E501
from odahuflow.sdk.models.environment_variable import EnvironmentVariable  # noqa: F401,E501
from odahuflow.sdk.models.model_identity import ModelIdentity  # noqa: F401,E501
from odahuflow.sdk.models.resource_requirements import ResourceRequirements  # noqa: F401,E501
from odahuflow.sdk.models import util


class ModelTrainingSpec(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, args: List[str]=None, data: List[DataBindingDir]=None, entrypoint: str=None, envs: List[EnvironmentVariable]=None, hyper_parameters: object=None, image: str=None, model: ModelIdentity=None, output_connection: str=None, reference: str=None, resources: ResourceRequirements=None, toolchain: str=None, vcs_name: str=None, work_dir: str=None):  # noqa: E501
        """ModelTrainingSpec - a model defined in Swagger

        :param args: The args of this ModelTrainingSpec.  # noqa: E501
        :type args: List[str]
        :param data: The data of this ModelTrainingSpec.  # noqa: E501
        :type data: List[DataBindingDir]
        :param entrypoint: The entrypoint of this ModelTrainingSpec.  # noqa: E501
        :type entrypoint: str
        :param envs: The envs of this ModelTrainingSpec.  # noqa: E501
        :type envs: List[EnvironmentVariable]
        :param hyper_parameters: The hyper_parameters of this ModelTrainingSpec.  # noqa: E501
        :type hyper_parameters: object
        :param image: The image of this ModelTrainingSpec.  # noqa: E501
        :type image: str
        :param model: The model of this ModelTrainingSpec.  # noqa: E501
        :type model: ModelIdentity
        :param output_connection: The output_connection of this ModelTrainingSpec.  # noqa: E501
        :type output_connection: str
        :param reference: The reference of this ModelTrainingSpec.  # noqa: E501
        :type reference: str
        :param resources: The resources of this ModelTrainingSpec.  # noqa: E501
        :type resources: ResourceRequirements
        :param toolchain: The toolchain of this ModelTrainingSpec.  # noqa: E501
        :type toolchain: str
        :param vcs_name: The vcs_name of this ModelTrainingSpec.  # noqa: E501
        :type vcs_name: str
        :param work_dir: The work_dir of this ModelTrainingSpec.  # noqa: E501
        :type work_dir: str
        """
        self.swagger_types = {
            'args': List[str],
            'data': List[DataBindingDir],
            'entrypoint': str,
            'envs': List[EnvironmentVariable],
            'hyper_parameters': object,
            'image': str,
            'model': ModelIdentity,
            'output_connection': str,
            'reference': str,
            'resources': ResourceRequirements,
            'toolchain': str,
            'vcs_name': str,
            'work_dir': str
        }

        self.attribute_map = {
            'args': 'args',
            'data': 'data',
            'entrypoint': 'entrypoint',
            'envs': 'envs',
            'hyper_parameters': 'hyperParameters',
            'image': 'image',
            'model': 'model',
            'output_connection': 'outputConnection',
            'reference': 'reference',
            'resources': 'resources',
            'toolchain': 'toolchain',
            'vcs_name': 'vcsName',
            'work_dir': 'workDir'
        }

        self._args = args
        self._data = data
        self._entrypoint = entrypoint
        self._envs = envs
        self._hyper_parameters = hyper_parameters
        self._image = image
        self._model = model
        self._output_connection = output_connection
        self._reference = reference
        self._resources = resources
        self._toolchain = toolchain
        self._vcs_name = vcs_name
        self._work_dir = work_dir

    @classmethod
    def from_dict(cls, dikt) -> 'ModelTrainingSpec':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ModelTrainingSpec of this ModelTrainingSpec.  # noqa: E501
        :rtype: ModelTrainingSpec
        """
        return util.deserialize_model(dikt, cls)

    @property
    def args(self) -> List[str]:
        """Gets the args of this ModelTrainingSpec.


        :return: The args of this ModelTrainingSpec.
        :rtype: List[str]
        """
        return self._args

    @args.setter
    def args(self, args: List[str]):
        """Sets the args of this ModelTrainingSpec.


        :param args: The args of this ModelTrainingSpec.
        :type args: List[str]
        """

        self._args = args

    @property
    def data(self) -> List[DataBindingDir]:
        """Gets the data of this ModelTrainingSpec.

        Input data for a training  # noqa: E501

        :return: The data of this ModelTrainingSpec.
        :rtype: List[DataBindingDir]
        """
        return self._data

    @data.setter
    def data(self, data: List[DataBindingDir]):
        """Sets the data of this ModelTrainingSpec.

        Input data for a training  # noqa: E501

        :param data: The data of this ModelTrainingSpec.
        :type data: List[DataBindingDir]
        """

        self._data = data

    @property
    def entrypoint(self) -> str:
        """Gets the entrypoint of this ModelTrainingSpec.

        Model training file. It can be python\\bash script or jupiter notebook  # noqa: E501

        :return: The entrypoint of this ModelTrainingSpec.
        :rtype: str
        """
        return self._entrypoint

    @entrypoint.setter
    def entrypoint(self, entrypoint: str):
        """Sets the entrypoint of this ModelTrainingSpec.

        Model training file. It can be python\\bash script or jupiter notebook  # noqa: E501

        :param entrypoint: The entrypoint of this ModelTrainingSpec.
        :type entrypoint: str
        """

        self._entrypoint = entrypoint

    @property
    def envs(self) -> List[EnvironmentVariable]:
        """Gets the envs of this ModelTrainingSpec.

        Custom environment variables that should be set before entrypoint invocation.  # noqa: E501

        :return: The envs of this ModelTrainingSpec.
        :rtype: List[EnvironmentVariable]
        """
        return self._envs

    @envs.setter
    def envs(self, envs: List[EnvironmentVariable]):
        """Sets the envs of this ModelTrainingSpec.

        Custom environment variables that should be set before entrypoint invocation.  # noqa: E501

        :param envs: The envs of this ModelTrainingSpec.
        :type envs: List[EnvironmentVariable]
        """

        self._envs = envs

    @property
    def hyper_parameters(self) -> object:
        """Gets the hyper_parameters of this ModelTrainingSpec.

        Model training hyperParameters in parameter:value format  # noqa: E501

        :return: The hyper_parameters of this ModelTrainingSpec.
        :rtype: object
        """
        return self._hyper_parameters

    @hyper_parameters.setter
    def hyper_parameters(self, hyper_parameters: object):
        """Sets the hyper_parameters of this ModelTrainingSpec.

        Model training hyperParameters in parameter:value format  # noqa: E501

        :param hyper_parameters: The hyper_parameters of this ModelTrainingSpec.
        :type hyper_parameters: object
        """

        self._hyper_parameters = hyper_parameters

    @property
    def image(self) -> str:
        """Gets the image of this ModelTrainingSpec.

        Train image  # noqa: E501

        :return: The image of this ModelTrainingSpec.
        :rtype: str
        """
        return self._image

    @image.setter
    def image(self, image: str):
        """Sets the image of this ModelTrainingSpec.

        Train image  # noqa: E501

        :param image: The image of this ModelTrainingSpec.
        :type image: str
        """

        self._image = image

    @property
    def model(self) -> ModelIdentity:
        """Gets the model of this ModelTrainingSpec.

        Model Identity  # noqa: E501

        :return: The model of this ModelTrainingSpec.
        :rtype: ModelIdentity
        """
        return self._model

    @model.setter
    def model(self, model: ModelIdentity):
        """Sets the model of this ModelTrainingSpec.

        Model Identity  # noqa: E501

        :param model: The model of this ModelTrainingSpec.
        :type model: ModelIdentity
        """

        self._model = model

    @property
    def output_connection(self) -> str:
        """Gets the output_connection of this ModelTrainingSpec.

        Name of Connection to storage where training output artifact will be stored. Permitted connection types are defined by specific toolchain  # noqa: E501

        :return: The output_connection of this ModelTrainingSpec.
        :rtype: str
        """
        return self._output_connection

    @output_connection.setter
    def output_connection(self, output_connection: str):
        """Sets the output_connection of this ModelTrainingSpec.

        Name of Connection to storage where training output artifact will be stored. Permitted connection types are defined by specific toolchain  # noqa: E501

        :param output_connection: The output_connection of this ModelTrainingSpec.
        :type output_connection: str
        """

        self._output_connection = output_connection

    @property
    def reference(self) -> str:
        """Gets the reference of this ModelTrainingSpec.

        VCS Reference  # noqa: E501

        :return: The reference of this ModelTrainingSpec.
        :rtype: str
        """
        return self._reference

    @reference.setter
    def reference(self, reference: str):
        """Sets the reference of this ModelTrainingSpec.

        VCS Reference  # noqa: E501

        :param reference: The reference of this ModelTrainingSpec.
        :type reference: str
        """

        self._reference = reference

    @property
    def resources(self) -> ResourceRequirements:
        """Gets the resources of this ModelTrainingSpec.

        Resources for model container The same format like k8s uses for pod resources.  # noqa: E501

        :return: The resources of this ModelTrainingSpec.
        :rtype: ResourceRequirements
        """
        return self._resources

    @resources.setter
    def resources(self, resources: ResourceRequirements):
        """Sets the resources of this ModelTrainingSpec.

        Resources for model container The same format like k8s uses for pod resources.  # noqa: E501

        :param resources: The resources of this ModelTrainingSpec.
        :type resources: ResourceRequirements
        """

        self._resources = resources

    @property
    def toolchain(self) -> str:
        """Gets the toolchain of this ModelTrainingSpec.

        IntegrationName of toolchain  # noqa: E501

        :return: The toolchain of this ModelTrainingSpec.
        :rtype: str
        """
        return self._toolchain

    @toolchain.setter
    def toolchain(self, toolchain: str):
        """Sets the toolchain of this ModelTrainingSpec.

        IntegrationName of toolchain  # noqa: E501

        :param toolchain: The toolchain of this ModelTrainingSpec.
        :type toolchain: str
        """

        self._toolchain = toolchain

    @property
    def vcs_name(self) -> str:
        """Gets the vcs_name of this ModelTrainingSpec.

        Name of Connection resource. Must exists  # noqa: E501

        :return: The vcs_name of this ModelTrainingSpec.
        :rtype: str
        """
        return self._vcs_name

    @vcs_name.setter
    def vcs_name(self, vcs_name: str):
        """Sets the vcs_name of this ModelTrainingSpec.

        Name of Connection resource. Must exists  # noqa: E501

        :param vcs_name: The vcs_name of this ModelTrainingSpec.
        :type vcs_name: str
        """

        self._vcs_name = vcs_name

    @property
    def work_dir(self) -> str:
        """Gets the work_dir of this ModelTrainingSpec.

        Directory with model scripts/files in a git repository  # noqa: E501

        :return: The work_dir of this ModelTrainingSpec.
        :rtype: str
        """
        return self._work_dir

    @work_dir.setter
    def work_dir(self, work_dir: str):
        """Sets the work_dir of this ModelTrainingSpec.

        Directory with model scripts/files in a git repository  # noqa: E501

        :param work_dir: The work_dir of this ModelTrainingSpec.
        :type work_dir: str
        """

        self._work_dir = work_dir



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models.connection_spec import ConnectionSpec  # noqa: F401,E501
from odahuflow.sdk.models.connection_status import ConnectionStatus  # noqa: F401,E501
from odahuflow.sdk.models import util


class Connection(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, id: str=None, spec: ConnectionSpec=None, status: ConnectionStatus=None):  # noqa: E501
        """Connection - a model defined in Swagger

        :param id: The id of this Connection.  # noqa: E501
        :type id: str
        :param spec: The spec of this Connection.  # noqa: E501
        :type spec: ConnectionSpec
        :param status: The status of this Connection.  # noqa: E501
        :type status: ConnectionStatus
        """
        self.swagger_types = {
            'id': str,
            'spec': ConnectionSpec,
            'status': ConnectionStatus
        }

        self.attribute_map = {
            'id': 'id',
            'spec': 'spec',
            'status': 'status'
        }

        self._id = id
        self._spec = spec
        self._status = status

    @classmethod
    def from_dict(cls, dikt) -> 'Connection':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Connection of this Connection.  # noqa: E501
        :rtype: Connection
        """
        return util.deserialize_model(dikt, cls)

    @property
    def id(self) -> str:
        """Gets the id of this Connection.

        Connection id  # noqa: E501

        :return: The id of this Connection.
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id: str):
        """Sets the id of this Connection.

        Connection id  # noqa: E501

        :param id: The id of this Connection.
        :type id: str
        """

        self._id = id

    @property
    def spec(self) -> ConnectionSpec:
        """Gets the spec of this Connection.

        Connection specification  # noqa: E501

        :return: The spec of this Connection.
        :rtype: ConnectionSpec
        """
        return self._spec

    @spec.setter
    def spec(self, spec: ConnectionSpec):
        """Sets the spec of this Connection.

        Connection specification  # noqa: E501

        :param spec: The spec of this Connection.
        :type spec: ConnectionSpec
        """

        self._spec = spec

    @property
    def status(self) -> ConnectionStatus:
        """Gets the status of this Connection.

        Connection status  # noqa: E501

        :return: The status of this Connection.
        :rtype: ConnectionStatus
        """
        return self._status

    @status.setter
    def status(self, status: ConnectionStatus):
        """Sets the status of this Connection.

        Connection status  # noqa: E501

        :param status: The status of this Connection.
        :type status: ConnectionStatus
        """

        self._status = status



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Model HTTP API client and utils
"""
import json
import logging

import requests
from requests.compat import urlencode
from requests.utils import to_key_val_list
from urllib3.exceptions import HTTPError

from odahuflow.sdk import config
from odahuflow.sdk.clients.route import ModelRouteClient
from odahuflow.sdk.utils import ensure_function_succeed

LOGGER = logging.getLogger(__name__)


def encode_http_params(data):
    """
    Encode HTTP parameters to URL query string

    :param data: data as text or tuple/list
    :type data: str or bytes or tuple or list
    :return: str -- encoded data
    """
    if isinstance(data, (str, bytes)):
        return urlencode(data)
    elif hasattr(data, '__iter__'):
        result = []
        for key, value in to_key_val_list(data):
            if value is not None:
                result.append(
                    (key.encode('utf-8') if isinstance(key, str) else key,
                     value.encode('utf-8') if isinstance(value, str) else value))
        return urlencode(result, doseq=True)
    raise ValueError('Invalid argument')


def calculate_url(host: str, url: str = None, model_route: str = None, model_deployment: str = None,
                  url_prefix: str = None, mr_client: ModelRouteClient = None):
    """
    Calculate url for model

    :param host: edge host
    :param url: full url to model api
    :param model_route: model route name
    :param model_deployment: model deployment name
    :param url_prefix: model prefix
    :return: model url
    """
    if url:
        return url

    if url_prefix:
        LOGGER.debug('')
        return f'{host}{url_prefix}'

    model_route = model_route or model_deployment
    if model_route:
        if mr_client is None:
            mr_client = ModelRouteClient()

        model_route = mr_client.get(model_route)

        LOGGER.debug('Found model route: %s', model_route)
        return model_route.status.edge_url

    raise NotImplementedError("Cannot create a model url")


def calculate_url_from_config():
    """
    Calculate url for model with config values

    :return: model url
    """
    return calculate_url(config.MODEL_HOST, config.MODEL_SERVER_URL, config.MODEL_ROUTE_NAME,
                         config.MODEL_DEPLOYMENT_NAME, config.MODEL_PREFIX_URL)


class ModelClient:
    """
    Model HTTP client
    """

    def __init__(self, url=None, token=None, http_client=requests,
                 http_exception=requests.exceptions.RequestException,
                 timeout=None):
        """
        Build client

        :param token: API token value to use (default: None)
        :type token: str
        :param http_client: HTTP client (default: requests)
        :type http_client: python class that implements requests-like post & get methods
        :param http_exception: http_client exception class, which can be thrown by http_client in case of some errors
        :type http_exception: python class that implements Exception class interface
        :param timeout: timeout for connections
        :type timeout: int
        """
        self._url = url
        self._token = token
        self._http_client = http_client
        self._http_exception = http_exception
        self._timeout = timeout

        LOGGER.debug('Model client params: %s, %s, %s, %s, %s', url, token, http_client, http_exception, timeout)

    @property
    def api_url(self):
        """
        Build API root URL

        :return: str -- api root url
        """
        return '{host}/api/model'.format(host=self._url)

    def build_batch_url(self, endpoint=None):
        """
        Build API batch invoke URL

        :param endpoint: (Optional) target endpoint
        :type endpoint: str
        :return: str -- batch invoke url
        """
        if endpoint:
            return '{}/batch/{}'.format(self.api_url, endpoint)
        return '{}/batch'.format(self.api_url)

    @property
    def info_url(self):
        """
        Build API info URL

        :return: str -- info url
        """
        return self.api_url + '/info'

    @staticmethod
    def _parse_response(response):
        """
        Parse model response (requests or FlaskClient)

        :param response: model HTTP response
        :type response: object with .text or .data and .status_code attributes
        :return: dict -- parsed response
        """
        data = response.text if hasattr(response, 'text') else response.data

        if isinstance(data, bytes):
            data = data.decode('utf-8')

        try:
            data = json.loads(data)
        except ValueError:
            pass

        if not 200 <= response.status_code < 400:
            url = response.url if hasattr(response, 'url') else None
            raise Exception('Wrong status code returned: {}. Data: {}. URL: {}'
                            .format(response.status_code, data, url))

        return data

    @property
    def _additional_kwargs(self):
        """
        Get additional HTTP client key-value arguments like timestamp

        :return: dict -- additional kwargs
        """
        kwargs = {}
        if self._token:
            kwargs['headers'] = {'Authorization': 'Bearer {token}'.format(token=self._token)}
        if self._timeout is not None:
            kwargs['timeout'] = self._timeout
        return kwargs

    def _request(self, http_method, url, data=None, files=None, retries=10, sleep=3, **kwargs):
        """
        Send request with provided method and other parameters
        :param http_method: HTTP method
        :type http_method: str
        :param url: url to send request to
        :type url: str
        :param data: request data
        :type data: any
        :param files: files to send with request
        :type files: dict
        :param retries: How many times to retry executing a request
        :type retries: int
        :param sleep: How much time to sleep between retries in case of errors
        :type sleep: int
        :return: dict -- parsed model response
        """
        http_method = http_method.lower()

        client_method = getattr(self._http_client, http_method)

        if data:
            kwargs['data'] = data
        if files:
            kwargs['files'] = files

        def check_function():
            try:
                return client_method(url, **kwargs)
            except (self._http_exception, HTTPError) as e:
                LOGGER.error('Failed to connect to {}: {}.'.format(url, e))

        response = ensure_function_succeed(check_function, retries, sleep)
        if response is None:
            raise self._http_exception('HTTP request failed')
        return self._parse_response(response)

    def invoke(self, **parameters):
        """
        Invoke model with parameters

        :param parameters: parameters for model
        :type parameters: dict[str, object] -- dictionary with parameters
        :return: dict -- parsed model response
        """
        return self._request('post', f'{self.api_url}/invoke', **self._additional_kwargs, json=parameters)

    def info(self):
        """
        Get model info

        :return: dict -- parsed model info
        """
        return self._request('get', self.info_url, **self._additional_kwargs)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class PackagingIntegrationStatus(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, created_at: str=None, updated_at: str=None):  # noqa: E501
        """PackagingIntegrationStatus - a model defined in Swagger

        :param created_at: The created_at of this PackagingIntegrationStatus.  # noqa: E501
        :type created_at: str
        :param updated_at: The updated_at of this PackagingIntegrationStatus.  # noqa: E501
        :type updated_at: str
        """
        self.swagger_types = {
            'created_at': str,
            'updated_at': str
        }

        self.attribute_map = {
            'created_at': 'createdAt',
            'updated_at': 'updatedAt'
        }

        self._created_at = created_at
        self._updated_at = updated_at

    @classmethod
    def from_dict(cls, dikt) -> 'PackagingIntegrationStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The PackagingIntegrationStatus of this PackagingIntegrationStatus.  # noqa: E501
        :rtype: PackagingIntegrationStatus
        """
        return util.deserialize_model(dikt, cls)

    @property
    def created_at(self) -> str:
        """Gets the created_at of this PackagingIntegrationStatus.


        :return: The created_at of this PackagingIntegrationStatus.
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at: str):
        """Sets the created_at of this PackagingIntegrationStatus.


        :param created_at: The created_at of this PackagingIntegrationStatus.
        :type created_at: str
        """

        self._created_at = created_at

    @property
    def updated_at(self) -> str:
        """Gets the updated_at of this PackagingIntegrationStatus.


        :return: The updated_at of this PackagingIntegrationStatus.
        :rtype: str
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at: str):
        """Sets the updated_at of this PackagingIntegrationStatus.


        :param updated_at: The updated_at of this PackagingIntegrationStatus.
        :type updated_at: str
        """

        self._updated_at = updated_at



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models.connection_spec import ConnectionSpec  # noqa: F401,E501
from odahuflow.sdk.models import util


class InputDataBindingDir(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, data_binding: ConnectionSpec=None, local_path: str=None, remote_path: str=None):  # noqa: E501
        """InputDataBindingDir - a model defined in Swagger

        :param data_binding: The data_binding of this InputDataBindingDir.  # noqa: E501
        :type data_binding: ConnectionSpec
        :param local_path: The local_path of this InputDataBindingDir.  # noqa: E501
        :type local_path: str
        :param remote_path: The remote_path of this InputDataBindingDir.  # noqa: E501
        :type remote_path: str
        """
        self.swagger_types = {
            'data_binding': ConnectionSpec,
            'local_path': str,
            'remote_path': str
        }

        self.attribute_map = {
            'data_binding': 'dataBinding',
            'local_path': 'localPath',
            'remote_path': 'remotePath'
        }

        self._data_binding = data_binding
        self._local_path = local_path
        self._remote_path = remote_path

    @classmethod
    def from_dict(cls, dikt) -> 'InputDataBindingDir':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The InputDataBindingDir of this InputDataBindingDir.  # noqa: E501
        :rtype: InputDataBindingDir
        """
        return util.deserialize_model(dikt, cls)

    @property
    def data_binding(self) -> ConnectionSpec:
        """Gets the data_binding of this InputDataBindingDir.

        Connection specific for data  # noqa: E501

        :return: The data_binding of this InputDataBindingDir.
        :rtype: ConnectionSpec
        """
        return self._data_binding

    @data_binding.setter
    def data_binding(self, data_binding: ConnectionSpec):
        """Sets the data_binding of this InputDataBindingDir.

        Connection specific for data  # noqa: E501

        :param data_binding: The data_binding of this InputDataBindingDir.
        :type data_binding: ConnectionSpec
        """

        self._data_binding = data_binding

    @property
    def local_path(self) -> str:
        """Gets the local_path of this InputDataBindingDir.

        Local path  # noqa: E501

        :return: The local_path of this InputDataBindingDir.
        :rtype: str
        """
        return self._local_path

    @local_path.setter
    def local_path(self, local_path: str):
        """Sets the local_path of this InputDataBindingDir.

        Local path  # noqa: E501

        :param local_path: The local_path of this InputDataBindingDir.
        :type local_path: str
        """

        self._local_path = local_path

    @property
    def remote_path(self) -> str:
        """Gets the remote_path of this InputDataBindingDir.

        Remote path  # noqa: E501

        :return: The remote_path of this InputDataBindingDir.
        :rtype: str
        """
        return self._remote_path

    @remote_path.setter
    def remote_path(self, remote_path: str):
        """Sets the remote_path of this InputDataBindingDir.

        Remote path  # noqa: E501

        :param remote_path: The remote_path of this InputDataBindingDir.
        :type remote_path: str
        """

        self._remote_path = remote_path



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class ConnectionStatus(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, created_at: str=None, secret_name: str=None, service_account: str=None, updated_at: str=None):  # noqa: E501
        """ConnectionStatus - a model defined in Swagger

        :param created_at: The created_at of this ConnectionStatus.  # noqa: E501
        :type created_at: str
        :param secret_name: The secret_name of this ConnectionStatus.  # noqa: E501
        :type secret_name: str
        :param service_account: The service_account of this ConnectionStatus.  # noqa: E501
        :type service_account: str
        :param updated_at: The updated_at of this ConnectionStatus.  # noqa: E501
        :type updated_at: str
        """
        self.swagger_types = {
            'created_at': str,
            'secret_name': str,
            'service_account': str,
            'updated_at': str
        }

        self.attribute_map = {
            'created_at': 'createdAt',
            'secret_name': 'secretName',
            'service_account': 'serviceAccount',
            'updated_at': 'updatedAt'
        }

        self._created_at = created_at
        self._secret_name = secret_name
        self._service_account = service_account
        self._updated_at = updated_at

    @classmethod
    def from_dict(cls, dikt) -> 'ConnectionStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ConnectionStatus of this ConnectionStatus.  # noqa: E501
        :rtype: ConnectionStatus
        """
        return util.deserialize_model(dikt, cls)

    @property
    def created_at(self) -> str:
        """Gets the created_at of this ConnectionStatus.


        :return: The created_at of this ConnectionStatus.
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at: str):
        """Sets the created_at of this ConnectionStatus.


        :param created_at: The created_at of this ConnectionStatus.
        :type created_at: str
        """

        self._created_at = created_at

    @property
    def secret_name(self) -> str:
        """Gets the secret_name of this ConnectionStatus.

        Kubernetes secret name  # noqa: E501

        :return: The secret_name of this ConnectionStatus.
        :rtype: str
        """
        return self._secret_name

    @secret_name.setter
    def secret_name(self, secret_name: str):
        """Sets the secret_name of this ConnectionStatus.

        Kubernetes secret name  # noqa: E501

        :param secret_name: The secret_name of this ConnectionStatus.
        :type secret_name: str
        """

        self._secret_name = secret_name

    @property
    def service_account(self) -> str:
        """Gets the service_account of this ConnectionStatus.

        Kubernetes service account  # noqa: E501

        :return: The service_account of this ConnectionStatus.
        :rtype: str
        """
        return self._service_account

    @service_account.setter
    def service_account(self, service_account: str):
        """Sets the service_account of this ConnectionStatus.

        Kubernetes service account  # noqa: E501

        :param service_account: The service_account of this ConnectionStatus.
        :type service_account: str
        """

        self._service_account = service_account

    @property
    def updated_at(self) -> str:
        """Gets the updated_at of this ConnectionStatus.


        :return: The updated_at of this ConnectionStatus.
        :rtype: str
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at: str):
        """Sets the updated_at of this ConnectionStatus.


        :param updated_at: The updated_at of this ConnectionStatus.
        :type updated_at: str
        """

        self._updated_at = updated_at



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class TrainingConfiguration(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, metric_url: str=None):  # noqa: E501
        """TrainingConfiguration - a model defined in Swagger

        :param metric_url: The metric_url of this TrainingConfiguration.  # noqa: E501
        :type metric_url: str
        """
        self.swagger_types = {
            'metric_url': str
        }

        self.attribute_map = {
            'metric_url': 'metricUrl'
        }

        self._metric_url = metric_url

    @classmethod
    def from_dict(cls, dikt) -> 'TrainingConfiguration':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The TrainingConfiguration of this TrainingConfiguration.  # noqa: E501
        :rtype: TrainingConfiguration
        """
        return util.deserialize_model(dikt, cls)

    @property
    def metric_url(self) -> str:
        """Gets the metric_url of this TrainingConfiguration.


        :return: The metric_url of this TrainingConfiguration.
        :rtype: str
        """
        return self._metric_url

    @metric_url.setter
    def metric_url(self, metric_url: str):
        """Sets the metric_url of this TrainingConfiguration.


        :param metric_url: The metric_url of this TrainingConfiguration.
        :type metric_url: str
        """

        self._metric_url = metric_url



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

"""
Helpers for working with `Trained Model Binary`
Docs: https://odahu.github.io/gen_glossary.html#term-trained-model-binary
"""
import json
import logging
import os
import uuid

import pydantic
import yaml
from odahuflow.sdk.gppi import entrypoint_invoke
from odahuflow.sdk.gppi.models import OdahuflowProjectManifest, OdahuflowProjectManifestBinaries
from odahuflow.sdk.io_proc_utils import run

_logger = logging.getLogger(__name__)

PROJECT_FILE = 'odahuflow.project.yaml'

VALIDATION_FAILED_EXCEPTION_MESSAGE = 'Exception raised while invoke model entrypoint. ' \
                                      'GPPI Self check is failed. ' \
                                      'Fix your model and try again.'


def _get_conda_bin_executable(executable_name):
    """
    Return path to the specified executable, assumed to be discoverable within the 'bin'
    subdirectory of a conda installation.
    """

    if "CONDA_EXE" in os.environ:
        conda_bin_dir = os.path.dirname(os.environ["CONDA_EXE"])
        executable = os.path.join(conda_bin_dir, executable_name)
    else:
        executable = executable_name
    return executable


def _get_activate_conda_command(conda_env_name):
    #  Checking for newer conda versions
    if 'CONDA_EXE' in os.environ:
        conda_path = _get_conda_bin_executable("conda")
        activate_conda_env = ['source ' + os.path.dirname(conda_path) +
                              '/../etc/profile.d/conda.sh']
        activate_conda_env += ["conda activate {0} 1>&2".format(conda_env_name)]
    else:
        activate_path = _get_conda_bin_executable("activate")
        # in case os name is not 'nt', we are not running on windows. It introduces
        # bash command otherwise.
        if os.name != "nt":
            activate_conda_env = ["source %s %s 1>&2" % (activate_path, conda_env_name)]
        else:
            activate_conda_env = ["conda %s %s 1>&2" % (activate_path, conda_env_name)]
    return ' && '.join(activate_conda_env)


def _check_conda():
    conda_path = _get_conda_bin_executable("conda")
    try:
        run(conda_path, "--help", stream_output=False)
    except Exception as exc_info:
        raise Exception("Could not find Conda executable at {0}. "
                        "Ensure Conda is installed as per the instructions "
                        "at https://conda.io/docs/user-guide/install/index.html") from exc_info


def load_odahuflow_project_manifest(manifest_path) -> OdahuflowProjectManifest:
    """
    Extract model manifest from file to object

    :param model: path to unpacked model (folder)
    :type model: str
    :return: None
    """
    manifest_file = os.path.join(manifest_path, PROJECT_FILE)
    if not manifest_file:
        raise Exception(f'Can not find manifest file {manifest_file}')

    with open(manifest_file, 'r') as manifest_stream:
        data = manifest_stream.read()

        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            try:
                data = yaml.safe_load(data)
            except yaml.YAMLError as decode_error:
                raise ValueError(f'Cannot decode ModelPacking resource file: {decode_error}')

        try:
            return OdahuflowProjectManifest(**data)
        except pydantic.ValidationError as valid_error:
            raise Exception(f'Legion manifest file is in incorrect format: {valid_error}')


class ExecutionEnvironment:

    def __init__(self, name: str, binaries: OdahuflowProjectManifestBinaries, manifest_path: str, skip_deps=False):

        self.check_environment_executables()

        if not name:
            name = str(uuid.uuid4())

        self._name = name
        self._binaries = binaries
        self._manifest_path = manifest_path

        self.create_env()

        if skip_deps:
            _logger.warning(f'Flag "skip_deps"=True. Installing deps is skipped')
        else:
            _logger.info(f'Start to install dependencies for {self.__class__}')
            self.install_dependencies()

    def __str__(self):
        return f'{self.__class__}: name: {self._name}'

    @property
    def binaries(self):
        return self._binaries

    def execute(self, command: str, cwd: str = None, stream_output: bool = True):
        raise NotImplementedError

    def install_dependencies(self):
        raise NotImplementedError

    def create_env(self):
        raise NotImplementedError

    def check_environment_executables(self):
        raise NotImplementedError


class CondaExecutionEnvironment(ExecutionEnvironment):

    def check_environment_executables(self):
        _check_conda()
        _logger.info('Conda environment executables are detected')

    def env_created(self):
        conda_exec = _get_conda_bin_executable('conda')
        _1, stdout, _2 = run(conda_exec, 'env', 'list', '--json', stream_output=False)
        env_names = [os.path.basename(env) for env in json.loads(stdout)['envs']]
        return self._name in env_names

    def create_env(self):
        env_id = self._name
        conda_exec = _get_conda_bin_executable('conda')

        if self.env_created():
            _logger.info(f'Conda env with name {env_id!r} already exists. Creating step is skipped')
        else:
            _logger.info(f'Creating conda env with name {env_id!r}')
            run(conda_exec, 'create', '--yes', '--name', env_id, stream_output=False)

    def install_dependencies(self):

        env_id = self._name
        conda_exec = _get_conda_bin_executable('conda')

        # Install requirements from dep. list
        conda_dep_list = os.path.join(self._manifest_path, self.binaries.conda_path)
        _logger.info(f'Installing mandatory requirements from {conda_dep_list} to {env_id!r}')
        run(conda_exec, 'env', 'update', f'--name={env_id}', f'--file={conda_dep_list}', stream_output=False)

    def execute(self, command: str, cwd: str = None, stream_output: bool = True):
        activate_cmd = _get_activate_conda_command(self._name)
        _logger.info(f'cwd: {cwd}')
        return run('bash', '-c', f'{activate_cmd} && {command}', cwd=cwd, stream_output=False)


class GPPITrainedModelBinary:
    """
    Class that allows interact with TrainedModelBinary stored in GPPI interface
        * Validate `TrainedModelBinary`
        * Install `TrainedModelBinary` dependencies
        * Run self-check `TrainedModelBinary`

    Additional Info:
        * https://odahu.github.io/gen_glossary.html#term-general-python-prediction-interface
    """

    def __init__(self, manifest_path: str, use_current_env: bool = False, env_name: str = '', skip_deps=False):
        """

        :param manifest_path: path to dir where odahuflow.project.yaml
        :param use_current_env: if False, new python env will be created and all commands
        """
        self.manifest_path: str = manifest_path
        try:
            self.manifest: OdahuflowProjectManifest = load_odahuflow_project_manifest(manifest_path)
        except Exception as e:
            raise Exception(VALIDATION_FAILED_EXCEPTION_MESSAGE) from e
        self.use_current_env = use_current_env
        self.skip_deps = skip_deps

        self.exec_env = None
        if not use_current_env:
            _logger.info(f'Start initializing environment')
            self.exec_env: ExecutionEnvironment = self._init_exec_env(env_name)

    @property
    def model_dir(self):
        return os.path.join(self.manifest_path, self.manifest.model.workDir)

    def execute(self, command: str, cwd: str = None, stream_output: bool = True):
        if self.use_current_env:
            _logger.info(f'use_current_env flag = True. Start to execute command without changing environment')
            exit_code, output, err_ = run('bash', '-c', command, cwd=cwd, stream_output=stream_output)
        else:
            _logger.info(f'use_current_env flag = False. Start to execute command in {self.exec_env}')
            exit_code, output, err_ = self.exec_env.execute(command, cwd=cwd, stream_output=stream_output)

        _logger.info('Subprocess result: %s\n\nSTDOUT:\n%s\n\nSTDERR:%s\n ======' %
                     (exit_code, output, err_))

        return output

    def self_check(self):
        """
        Try to invoke public API of GPPI
        1) Invoke entrypoint.init()
        2) Invoke entrypoint.info()
        3) If input samples is found in package run entrypoint.predict_on_matrix()
        :return:
        """

        _logger.info('Start procedure of self checking for GPPI')
        try:
            self.execute(
                f'python  {entrypoint_invoke.__file__} -v --model {self.model_dir} '
                f'--entrypoint {self.manifest.model.entrypoint} self_check',
                stream_output=False
            )
        except Exception as e:
            raise Exception(VALIDATION_FAILED_EXCEPTION_MESSAGE) from e

        _logger.info('Self check is successful. GPPI model is packed as expected.')

    def predict(self, input_file: str, output_dir: str, output_file_name):
        """
        Do predictions
        :param input_file: JSON file with data for predictions
        :param output_dir: output directory where results will be saved
        :return:
        """
        try:
            self.execute(
                f'python  {entrypoint_invoke.__file__} -v --model {self.model_dir} '
                f'--entrypoint {self.manifest.model.entrypoint} predict {input_file} {output_dir} '
                f'{"--output_file_name "+output_file_name if output_file_name else ""}',
                stream_output=False
            )
        except Exception as e:
            raise Exception(VALIDATION_FAILED_EXCEPTION_MESSAGE) from e

        _logger.info('Prediction successfully completed')

    def info(self) -> str:
        """
        Show input/output model schema
        :return:
        """
        try:
            out = self.execute(
                f'python  {entrypoint_invoke.__file__} -v --model {self.model_dir} '
                f'--entrypoint {self.manifest.model.entrypoint} info',
                stream_output=False
            )
        except Exception as e:
            raise Exception(VALIDATION_FAILED_EXCEPTION_MESSAGE) from e

        return out

    def _init_exec_env(self, env_name: str) -> ExecutionEnvironment:
        if self.manifest.binaries.dependencies == 'conda':
            env = CondaExecutionEnvironment(env_name, self.manifest.binaries, self.manifest_path, self.skip_deps)
        else:
            raise RuntimeError(f'{self.manifest.binaries.dependencies} dependency manager is not recognized')

        return env



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class ModelDeploymentTarget(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, md_name: str=None, weight: int=None):  # noqa: E501
        """ModelDeploymentTarget - a model defined in Swagger

        :param md_name: The md_name of this ModelDeploymentTarget.  # noqa: E501
        :type md_name: str
        :param weight: The weight of this ModelDeploymentTarget.  # noqa: E501
        :type weight: int
        """
        self.swagger_types = {
            'md_name': str,
            'weight': int
        }

        self.attribute_map = {
            'md_name': 'mdName',
            'weight': 'weight'
        }

        self._md_name = md_name
        self._weight = weight

    @classmethod
    def from_dict(cls, dikt) -> 'ModelDeploymentTarget':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ModelDeploymentTarget of this ModelDeploymentTarget.  # noqa: E501
        :rtype: ModelDeploymentTarget
        """
        return util.deserialize_model(dikt, cls)

    @property
    def md_name(self) -> str:
        """Gets the md_name of this ModelDeploymentTarget.

        Model Deployment name  # noqa: E501

        :return: The md_name of this ModelDeploymentTarget.
        :rtype: str
        """
        return self._md_name

    @md_name.setter
    def md_name(self, md_name: str):
        """Sets the md_name of this ModelDeploymentTarget.

        Model Deployment name  # noqa: E501

        :param md_name: The md_name of this ModelDeploymentTarget.
        :type md_name: str
        """

        self._md_name = md_name

    @property
    def weight(self) -> int:
        """Gets the weight of this ModelDeploymentTarget.

        The proportion of traffic to be forwarded to the Model Deployment.  # noqa: E501

        :return: The weight of this ModelDeploymentTarget.
        :rtype: int
        """
        return self._weight

    @weight.setter
    def weight(self, weight: int):
        """Sets the weight of this ModelDeploymentTarget.

        The proportion of traffic to be forwarded to the Model Deployment.  # noqa: E501

        :param weight: The weight of this ModelDeploymentTarget.
        :type weight: int
        """

        self._weight = weight



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import logging
import os.path
import subprocess

LOGGER = logging.getLogger(__name__)


def setup_logging(verbose: bool = False) -> None:
    """
    Setup logging instance

    :param verbose: use verbose output
    :type verbose: bool
    """
    log_level = logging.DEBUG if verbose else logging.INFO

    logging.basicConfig(format='[odahuflow][%(levelname)5s] %(asctime)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S',
                        level=log_level)


def run(*args: str, cwd=None, stream_output: bool = True):
    """
    Run system command and stream / capture stdout and stderr

    :param args: Commands to run (list). E.g. ['cp, '-R', '/var/www', '/www']
    :type args: :py:class:`typing.List[str]`
    :param cwd: (Optional). Program working directory. Current is used if not provided.
    :type cwd: str
    :param stream_output: (Optional). Flag that enables streaming child process output to stdout and stderr.
    :return: typing.Union[int, typing.Tuple[int, str, str]] -- exit_code (for stream_output mode)
             or exit_code + stdout + stderr.
    """
    args_line = ' '.join(args)
    logging.info(f'Running command "{args_line}"')

    cmd_env = os.environ.copy()
    if stream_output:
        child = subprocess.Popen(args, env=cmd_env, cwd=cwd, universal_newlines=True,
                                 stdin=subprocess.PIPE)
        exit_code = child.wait()
        if exit_code != 0:
            raise Exception("Non-zero exitcode: %s" % exit_code)
        return exit_code
    else:
        child = subprocess.Popen(
            args, env=cmd_env, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE,
            cwd=cwd, universal_newlines=True)
        (stdout, stderr) = child.communicate()
        exit_code = child.wait()
        if exit_code != 0:
            raise Exception("Non-zero exit code: %s\n\nSTDOUT:\n%s\n\nSTDERR:%s\n ======" %
                            (exit_code, stdout, stderr))
        return exit_code, stdout, stderr



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
List of templates
"""
import os
import typing

_TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), 'examples')
_EXAMPLE_SUFFIX = ".odahu.yaml"


def get_odahuflow_template_names() -> typing.Set[str]:
    """
    The SDK package contains prebuilt examples of odahuflow entities
    This function return their names
    :return: template names
    """
    return {
        # Remove suffix from name
        f[:-len(_EXAMPLE_SUFFIX)] for _, _, filenames in os.walk(_TEMPLATE_DIR) for f in filenames
    }


def get_odahuflow_template_content(template_name: str) -> str:
    """
    Returns content of a template by name
    :param template_name: template name
    :return: template content
    """
    full_file_name = os.path.join(_TEMPLATE_DIR, template_name + _EXAMPLE_SUFFIX)
    if not os.path.exists(full_file_name):
        raise ValueError(f'Cannot find {template_name} template')

    with open(full_file_name, 'r', encoding='utf8') as fp:
        return fp.read()



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import os
import re

from setuptools import find_namespace_packages, setup

PACKAGE_ROOT_PATH = os.path.dirname(os.path.abspath(__file__))
VERSION_FILE = os.path.join(PACKAGE_ROOT_PATH, 'odahuflow/sdk', 'version.py')


def extract_version() -> str:
    """
    Extract version from .py file using regex

    :return: odahuflow version
    """
    with open(VERSION_FILE, 'rt') as version_file:
        file_content = version_file.read()
        VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
        mo = re.search(VSRE, file_content, re.M)
        if mo:
            return mo.group(1)
        else:
            raise RuntimeError("Unable to find version string in %s." % (file_content,))


with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='odahu-flow-sdk',
    version=extract_version(),
    description='Odahu-flow SDK',
    packages=find_namespace_packages(),
    url='https://github.com/odahu/odahu-flow',
    author='Vlad Tokarev, Vitalik Solodilov',
    author_email='vlad.tokarev.94@gmail.com, mcdkr@yandex.ru',
    license='Apache v2',
    install_requires=requirements,
)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models.connection import Connection  # noqa: F401,E501
from odahuflow.sdk.models.input_data_binding_dir import InputDataBindingDir  # noqa: F401,E501
from odahuflow.sdk.models.model_training import ModelTraining  # noqa: F401,E501
from odahuflow.sdk.models.toolchain_integration import ToolchainIntegration  # noqa: F401,E501
from odahuflow.sdk.models import util


class K8sTrainer(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, input_data: List[InputDataBindingDir]=None, model_training: ModelTraining=None, output_conn: Connection=None, toolchain_integration: ToolchainIntegration=None, vcs: Connection=None):  # noqa: E501
        """K8sTrainer - a model defined in Swagger

        :param input_data: The input_data of this K8sTrainer.  # noqa: E501
        :type input_data: List[InputDataBindingDir]
        :param model_training: The model_training of this K8sTrainer.  # noqa: E501
        :type model_training: ModelTraining
        :param output_conn: The output_conn of this K8sTrainer.  # noqa: E501
        :type output_conn: Connection
        :param toolchain_integration: The toolchain_integration of this K8sTrainer.  # noqa: E501
        :type toolchain_integration: ToolchainIntegration
        :param vcs: The vcs of this K8sTrainer.  # noqa: E501
        :type vcs: Connection
        """
        self.swagger_types = {
            'input_data': List[InputDataBindingDir],
            'model_training': ModelTraining,
            'output_conn': Connection,
            'toolchain_integration': ToolchainIntegration,
            'vcs': Connection
        }

        self.attribute_map = {
            'input_data': 'inputData',
            'model_training': 'modelTraining',
            'output_conn': 'outputConn',
            'toolchain_integration': 'toolchainIntegration',
            'vcs': 'vcs'
        }

        self._input_data = input_data
        self._model_training = model_training
        self._output_conn = output_conn
        self._toolchain_integration = toolchain_integration
        self._vcs = vcs

    @classmethod
    def from_dict(cls, dikt) -> 'K8sTrainer':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The K8sTrainer of this K8sTrainer.  # noqa: E501
        :rtype: K8sTrainer
        """
        return util.deserialize_model(dikt, cls)

    @property
    def input_data(self) -> List[InputDataBindingDir]:
        """Gets the input_data of this K8sTrainer.

        Connection for training data  # noqa: E501

        :return: The input_data of this K8sTrainer.
        :rtype: List[InputDataBindingDir]
        """
        return self._input_data

    @input_data.setter
    def input_data(self, input_data: List[InputDataBindingDir]):
        """Sets the input_data of this K8sTrainer.

        Connection for training data  # noqa: E501

        :param input_data: The input_data of this K8sTrainer.
        :type input_data: List[InputDataBindingDir]
        """

        self._input_data = input_data

    @property
    def model_training(self) -> ModelTraining:
        """Gets the model_training of this K8sTrainer.

        Model training  # noqa: E501

        :return: The model_training of this K8sTrainer.
        :rtype: ModelTraining
        """
        return self._model_training

    @model_training.setter
    def model_training(self, model_training: ModelTraining):
        """Sets the model_training of this K8sTrainer.

        Model training  # noqa: E501

        :param model_training: The model_training of this K8sTrainer.
        :type model_training: ModelTraining
        """

        self._model_training = model_training

    @property
    def output_conn(self) -> Connection:
        """Gets the output_conn of this K8sTrainer.

        Connection for trained model artifact  # noqa: E501

        :return: The output_conn of this K8sTrainer.
        :rtype: Connection
        """
        return self._output_conn

    @output_conn.setter
    def output_conn(self, output_conn: Connection):
        """Sets the output_conn of this K8sTrainer.

        Connection for trained model artifact  # noqa: E501

        :param output_conn: The output_conn of this K8sTrainer.
        :type output_conn: Connection
        """

        self._output_conn = output_conn

    @property
    def toolchain_integration(self) -> ToolchainIntegration:
        """Gets the toolchain_integration of this K8sTrainer.

        Toolchain integration  # noqa: E501

        :return: The toolchain_integration of this K8sTrainer.
        :rtype: ToolchainIntegration
        """
        return self._toolchain_integration

    @toolchain_integration.setter
    def toolchain_integration(self, toolchain_integration: ToolchainIntegration):
        """Sets the toolchain_integration of this K8sTrainer.

        Toolchain integration  # noqa: E501

        :param toolchain_integration: The toolchain_integration of this K8sTrainer.
        :type toolchain_integration: ToolchainIntegration
        """

        self._toolchain_integration = toolchain_integration

    @property
    def vcs(self) -> Connection:
        """Gets the vcs of this K8sTrainer.

        Connection for source code  # noqa: E501

        :return: The vcs of this K8sTrainer.
        :rtype: Connection
        """
        return self._vcs

    @vcs.setter
    def vcs(self, vcs: Connection):
        """Sets the vcs of this K8sTrainer.

        Connection for source code  # noqa: E501

        :param vcs: The vcs of this K8sTrainer.
        :type vcs: Connection
        """

        self._vcs = vcs



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
odahuflow utils functional
"""
import inspect
import logging
import time

from jinja2 import Environment, PackageLoader, select_autoescape

LOGGER = logging.getLogger(__name__)


def render_template(template_name, values=None):
    """
    Render template with parameters

    :param template_name: name of template without path (all templates should be placed in
                          odahuflow.templates directory)
    :param values: dict template variables or None
    :return: str rendered template
    """
    env = Environment(
        loader=PackageLoader(__name__, 'templates'),
        autoescape=select_autoescape(['tmpl'])
    )

    if not values:
        values = {}

    template = env.get_template(template_name)
    return template.render(values)



def get_function_description(callable_object):
    """
    Gather information about callable object to string

    :param callable_object: callable object to analyze
    :type callable_object: Callable[[], any]
    :return: str -- object description
    """
    object_class_name = callable_object.__class__.__name__
    if not callable(callable_object):
        return '<not callable object: {}>'.format(object_class_name)

    object_name = callable_object.__name__
    module_name = inspect.getmodule(callable_object)
    return '<{} {} in {}>'.format(object_class_name, object_name, module_name)


def ensure_function_succeed(function_to_call, retries, timeout, boolean_check=False):
    """
    Try to call function till it will return not None object.
    Raise if there are no retries left

    :param function_to_call: function to be called
    :type function_to_call: Callable[[], any]
    :param retries: count of retries
    :type retries: int
    :param timeout: timeout between retries
    :type timeout: int
    :param boolean_check: (Optional) check function for True-able value (by default value is checked for not None value)
    :type boolean_check: bool
    :return: Any -- result of successful function call or None if no retries left
    """
    function_description = get_function_description(function_to_call)
    for no in range(retries):
        LOGGER.debug('Calling {}'.format(function_description))
        result = function_to_call()

        if boolean_check:
            if result:
                return result
        else:
            if result is not None:
                return result

        if no < retries:
            LOGGER.debug('Retry {}/{} was failed'.format(no + 1, retries))
        if no < retries - 1:
            LOGGER.debug('Waiting {}s before next retry analysis'.format(timeout))
            time.sleep(timeout)

    LOGGER.error('No retries left for function {}'.format(function_description))
    return False if boolean_check else None



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models.resource_requirements import ResourceRequirements  # noqa: F401,E501
from odahuflow.sdk.models import util


class ModelDeploymentSpec(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, annotations: object=None, image: str=None, image_pull_conn_id: str=None, liveness_probe_initial_delay: int=None, max_replicas: int=None, min_replicas: int=None, readiness_probe_initial_delay: int=None, resources: ResourceRequirements=None, role_name: str=None):  # noqa: E501
        """ModelDeploymentSpec - a model defined in Swagger

        :param annotations: The annotations of this ModelDeploymentSpec.  # noqa: E501
        :type annotations: object
        :param image: The image of this ModelDeploymentSpec.  # noqa: E501
        :type image: str
        :param image_pull_conn_id: The image_pull_conn_id of this ModelDeploymentSpec.  # noqa: E501
        :type image_pull_conn_id: str
        :param liveness_probe_initial_delay: The liveness_probe_initial_delay of this ModelDeploymentSpec.  # noqa: E501
        :type liveness_probe_initial_delay: int
        :param max_replicas: The max_replicas of this ModelDeploymentSpec.  # noqa: E501
        :type max_replicas: int
        :param min_replicas: The min_replicas of this ModelDeploymentSpec.  # noqa: E501
        :type min_replicas: int
        :param readiness_probe_initial_delay: The readiness_probe_initial_delay of this ModelDeploymentSpec.  # noqa: E501
        :type readiness_probe_initial_delay: int
        :param resources: The resources of this ModelDeploymentSpec.  # noqa: E501
        :type resources: ResourceRequirements
        :param role_name: The role_name of this ModelDeploymentSpec.  # noqa: E501
        :type role_name: str
        """
        self.swagger_types = {
            'annotations': object,
            'image': str,
            'image_pull_conn_id': str,
            'liveness_probe_initial_delay': int,
            'max_replicas': int,
            'min_replicas': int,
            'readiness_probe_initial_delay': int,
            'resources': ResourceRequirements,
            'role_name': str
        }

        self.attribute_map = {
            'annotations': 'annotations',
            'image': 'image',
            'image_pull_conn_id': 'imagePullConnID',
            'liveness_probe_initial_delay': 'livenessProbeInitialDelay',
            'max_replicas': 'maxReplicas',
            'min_replicas': 'minReplicas',
            'readiness_probe_initial_delay': 'readinessProbeInitialDelay',
            'resources': 'resources',
            'role_name': 'roleName'
        }

        self._annotations = annotations
        self._image = image
        self._image_pull_conn_id = image_pull_conn_id
        self._liveness_probe_initial_delay = liveness_probe_initial_delay
        self._max_replicas = max_replicas
        self._min_replicas = min_replicas
        self._readiness_probe_initial_delay = readiness_probe_initial_delay
        self._resources = resources
        self._role_name = role_name

    @classmethod
    def from_dict(cls, dikt) -> 'ModelDeploymentSpec':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ModelDeploymentSpec of this ModelDeploymentSpec.  # noqa: E501
        :rtype: ModelDeploymentSpec
        """
        return util.deserialize_model(dikt, cls)

    @property
    def annotations(self) -> object:
        """Gets the annotations of this ModelDeploymentSpec.

        Annotations for model pods.  # noqa: E501

        :return: The annotations of this ModelDeploymentSpec.
        :rtype: object
        """
        return self._annotations

    @annotations.setter
    def annotations(self, annotations: object):
        """Sets the annotations of this ModelDeploymentSpec.

        Annotations for model pods.  # noqa: E501

        :param annotations: The annotations of this ModelDeploymentSpec.
        :type annotations: object
        """

        self._annotations = annotations

    @property
    def image(self) -> str:
        """Gets the image of this ModelDeploymentSpec.

        Model Docker image  # noqa: E501

        :return: The image of this ModelDeploymentSpec.
        :rtype: str
        """
        return self._image

    @image.setter
    def image(self, image: str):
        """Sets the image of this ModelDeploymentSpec.

        Model Docker image  # noqa: E501

        :param image: The image of this ModelDeploymentSpec.
        :type image: str
        """

        self._image = image

    @property
    def image_pull_conn_id(self) -> str:
        """Gets the image_pull_conn_id of this ModelDeploymentSpec.

        If pulling of your image requires authorization, then you should specify the connection id  # noqa: E501

        :return: The image_pull_conn_id of this ModelDeploymentSpec.
        :rtype: str
        """
        return self._image_pull_conn_id

    @image_pull_conn_id.setter
    def image_pull_conn_id(self, image_pull_conn_id: str):
        """Sets the image_pull_conn_id of this ModelDeploymentSpec.

        If pulling of your image requires authorization, then you should specify the connection id  # noqa: E501

        :param image_pull_conn_id: The image_pull_conn_id of this ModelDeploymentSpec.
        :type image_pull_conn_id: str
        """

        self._image_pull_conn_id = image_pull_conn_id

    @property
    def liveness_probe_initial_delay(self) -> int:
        """Gets the liveness_probe_initial_delay of this ModelDeploymentSpec.

        Initial delay for liveness probe of model pod  # noqa: E501

        :return: The liveness_probe_initial_delay of this ModelDeploymentSpec.
        :rtype: int
        """
        return self._liveness_probe_initial_delay

    @liveness_probe_initial_delay.setter
    def liveness_probe_initial_delay(self, liveness_probe_initial_delay: int):
        """Sets the liveness_probe_initial_delay of this ModelDeploymentSpec.

        Initial delay for liveness probe of model pod  # noqa: E501

        :param liveness_probe_initial_delay: The liveness_probe_initial_delay of this ModelDeploymentSpec.
        :type liveness_probe_initial_delay: int
        """

        self._liveness_probe_initial_delay = liveness_probe_initial_delay

    @property
    def max_replicas(self) -> int:
        """Gets the max_replicas of this ModelDeploymentSpec.

        Maximum number of pods for model. By default the max replicas parameter equals 1.  # noqa: E501

        :return: The max_replicas of this ModelDeploymentSpec.
        :rtype: int
        """
        return self._max_replicas

    @max_replicas.setter
    def max_replicas(self, max_replicas: int):
        """Sets the max_replicas of this ModelDeploymentSpec.

        Maximum number of pods for model. By default the max replicas parameter equals 1.  # noqa: E501

        :param max_replicas: The max_replicas of this ModelDeploymentSpec.
        :type max_replicas: int
        """

        self._max_replicas = max_replicas

    @property
    def min_replicas(self) -> int:
        """Gets the min_replicas of this ModelDeploymentSpec.

        Minimum number of pods for model. By default the min replicas parameter equals 0.  # noqa: E501

        :return: The min_replicas of this ModelDeploymentSpec.
        :rtype: int
        """
        return self._min_replicas

    @min_replicas.setter
    def min_replicas(self, min_replicas: int):
        """Sets the min_replicas of this ModelDeploymentSpec.

        Minimum number of pods for model. By default the min replicas parameter equals 0.  # noqa: E501

        :param min_replicas: The min_replicas of this ModelDeploymentSpec.
        :type min_replicas: int
        """

        self._min_replicas = min_replicas

    @property
    def readiness_probe_initial_delay(self) -> int:
        """Gets the readiness_probe_initial_delay of this ModelDeploymentSpec.

        Initial delay for readiness probe of model pod  # noqa: E501

        :return: The readiness_probe_initial_delay of this ModelDeploymentSpec.
        :rtype: int
        """
        return self._readiness_probe_initial_delay

    @readiness_probe_initial_delay.setter
    def readiness_probe_initial_delay(self, readiness_probe_initial_delay: int):
        """Sets the readiness_probe_initial_delay of this ModelDeploymentSpec.

        Initial delay for readiness probe of model pod  # noqa: E501

        :param readiness_probe_initial_delay: The readiness_probe_initial_delay of this ModelDeploymentSpec.
        :type readiness_probe_initial_delay: int
        """

        self._readiness_probe_initial_delay = readiness_probe_initial_delay

    @property
    def resources(self) -> ResourceRequirements:
        """Gets the resources of this ModelDeploymentSpec.

        Resources for model deployment The same format like k8s uses for pod resources.  # noqa: E501

        :return: The resources of this ModelDeploymentSpec.
        :rtype: ResourceRequirements
        """
        return self._resources

    @resources.setter
    def resources(self, resources: ResourceRequirements):
        """Sets the resources of this ModelDeploymentSpec.

        Resources for model deployment The same format like k8s uses for pod resources.  # noqa: E501

        :param resources: The resources of this ModelDeploymentSpec.
        :type resources: ResourceRequirements
        """

        self._resources = resources

    @property
    def role_name(self) -> str:
        """Gets the role_name of this ModelDeploymentSpec.

        Initial delay for readiness probe of model pod  # noqa: E501

        :return: The role_name of this ModelDeploymentSpec.
        :rtype: str
        """
        return self._role_name

    @role_name.setter
    def role_name(self, role_name: str):
        """Sets the role_name of this ModelDeploymentSpec.

        Initial delay for readiness probe of model pod  # noqa: E501

        :param role_name: The role_name of this ModelDeploymentSpec.
        :type role_name: str
        """

        self._role_name = role_name



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class ConnectionSpec(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, description: str=None, key_id: str=None, key_secret: str=None, password: str=None, public_key: str=None, reference: str=None, region: str=None, role: str=None, type: str=None, uri: str=None, username: str=None, web_ui_link: str=None):  # noqa: E501
        """ConnectionSpec - a model defined in Swagger

        :param description: The description of this ConnectionSpec.  # noqa: E501
        :type description: str
        :param key_id: The key_id of this ConnectionSpec.  # noqa: E501
        :type key_id: str
        :param key_secret: The key_secret of this ConnectionSpec.  # noqa: E501
        :type key_secret: str
        :param password: The password of this ConnectionSpec.  # noqa: E501
        :type password: str
        :param public_key: The public_key of this ConnectionSpec.  # noqa: E501
        :type public_key: str
        :param reference: The reference of this ConnectionSpec.  # noqa: E501
        :type reference: str
        :param region: The region of this ConnectionSpec.  # noqa: E501
        :type region: str
        :param role: The role of this ConnectionSpec.  # noqa: E501
        :type role: str
        :param type: The type of this ConnectionSpec.  # noqa: E501
        :type type: str
        :param uri: The uri of this ConnectionSpec.  # noqa: E501
        :type uri: str
        :param username: The username of this ConnectionSpec.  # noqa: E501
        :type username: str
        :param web_ui_link: The web_ui_link of this ConnectionSpec.  # noqa: E501
        :type web_ui_link: str
        """
        self.swagger_types = {
            'description': str,
            'key_id': str,
            'key_secret': str,
            'password': str,
            'public_key': str,
            'reference': str,
            'region': str,
            'role': str,
            'type': str,
            'uri': str,
            'username': str,
            'web_ui_link': str
        }

        self.attribute_map = {
            'description': 'description',
            'key_id': 'keyID',
            'key_secret': 'keySecret',
            'password': 'password',
            'public_key': 'publicKey',
            'reference': 'reference',
            'region': 'region',
            'role': 'role',
            'type': 'type',
            'uri': 'uri',
            'username': 'username',
            'web_ui_link': 'webUILink'
        }

        self._description = description
        self._key_id = key_id
        self._key_secret = key_secret
        self._password = password
        self._public_key = public_key
        self._reference = reference
        self._region = region
        self._role = role
        self._type = type
        self._uri = uri
        self._username = username
        self._web_ui_link = web_ui_link

    @classmethod
    def from_dict(cls, dikt) -> 'ConnectionSpec':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ConnectionSpec of this ConnectionSpec.  # noqa: E501
        :rtype: ConnectionSpec
        """
        return util.deserialize_model(dikt, cls)

    @property
    def description(self) -> str:
        """Gets the description of this ConnectionSpec.

        Custom description  # noqa: E501

        :return: The description of this ConnectionSpec.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description: str):
        """Sets the description of this ConnectionSpec.

        Custom description  # noqa: E501

        :param description: The description of this ConnectionSpec.
        :type description: str
        """

        self._description = description

    @property
    def key_id(self) -> str:
        """Gets the key_id of this ConnectionSpec.

        Key ID  # noqa: E501

        :return: The key_id of this ConnectionSpec.
        :rtype: str
        """
        return self._key_id

    @key_id.setter
    def key_id(self, key_id: str):
        """Sets the key_id of this ConnectionSpec.

        Key ID  # noqa: E501

        :param key_id: The key_id of this ConnectionSpec.
        :type key_id: str
        """

        self._key_id = key_id

    @property
    def key_secret(self) -> str:
        """Gets the key_secret of this ConnectionSpec.

        SSH or service account secret  # noqa: E501

        :return: The key_secret of this ConnectionSpec.
        :rtype: str
        """
        return self._key_secret

    @key_secret.setter
    def key_secret(self, key_secret: str):
        """Sets the key_secret of this ConnectionSpec.

        SSH or service account secret  # noqa: E501

        :param key_secret: The key_secret of this ConnectionSpec.
        :type key_secret: str
        """

        self._key_secret = key_secret

    @property
    def password(self) -> str:
        """Gets the password of this ConnectionSpec.

        Password  # noqa: E501

        :return: The password of this ConnectionSpec.
        :rtype: str
        """
        return self._password

    @password.setter
    def password(self, password: str):
        """Sets the password of this ConnectionSpec.

        Password  # noqa: E501

        :param password: The password of this ConnectionSpec.
        :type password: str
        """

        self._password = password

    @property
    def public_key(self) -> str:
        """Gets the public_key of this ConnectionSpec.

        SSH public key  # noqa: E501

        :return: The public_key of this ConnectionSpec.
        :rtype: str
        """
        return self._public_key

    @public_key.setter
    def public_key(self, public_key: str):
        """Sets the public_key of this ConnectionSpec.

        SSH public key  # noqa: E501

        :param public_key: The public_key of this ConnectionSpec.
        :type public_key: str
        """

        self._public_key = public_key

    @property
    def reference(self) -> str:
        """Gets the reference of this ConnectionSpec.

        VCS reference  # noqa: E501

        :return: The reference of this ConnectionSpec.
        :rtype: str
        """
        return self._reference

    @reference.setter
    def reference(self, reference: str):
        """Sets the reference of this ConnectionSpec.

        VCS reference  # noqa: E501

        :param reference: The reference of this ConnectionSpec.
        :type reference: str
        """

        self._reference = reference

    @property
    def region(self) -> str:
        """Gets the region of this ConnectionSpec.

        AWS region or GCP project  # noqa: E501

        :return: The region of this ConnectionSpec.
        :rtype: str
        """
        return self._region

    @region.setter
    def region(self, region: str):
        """Sets the region of this ConnectionSpec.

        AWS region or GCP project  # noqa: E501

        :param region: The region of this ConnectionSpec.
        :type region: str
        """

        self._region = region

    @property
    def role(self) -> str:
        """Gets the role of this ConnectionSpec.

        Service account role  # noqa: E501

        :return: The role of this ConnectionSpec.
        :rtype: str
        """
        return self._role

    @role.setter
    def role(self, role: str):
        """Sets the role of this ConnectionSpec.

        Service account role  # noqa: E501

        :param role: The role of this ConnectionSpec.
        :type role: str
        """

        self._role = role

    @property
    def type(self) -> str:
        """Gets the type of this ConnectionSpec.

        Required value. Available values:   * s3   * gcs   * azureblob   * git   * docker  # noqa: E501

        :return: The type of this ConnectionSpec.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type: str):
        """Sets the type of this ConnectionSpec.

        Required value. Available values:   * s3   * gcs   * azureblob   * git   * docker  # noqa: E501

        :param type: The type of this ConnectionSpec.
        :type type: str
        """

        self._type = type

    @property
    def uri(self) -> str:
        """Gets the uri of this ConnectionSpec.

        URI. It is required value  # noqa: E501

        :return: The uri of this ConnectionSpec.
        :rtype: str
        """
        return self._uri

    @uri.setter
    def uri(self, uri: str):
        """Sets the uri of this ConnectionSpec.

        URI. It is required value  # noqa: E501

        :param uri: The uri of this ConnectionSpec.
        :type uri: str
        """

        self._uri = uri

    @property
    def username(self) -> str:
        """Gets the username of this ConnectionSpec.

        Username  # noqa: E501

        :return: The username of this ConnectionSpec.
        :rtype: str
        """
        return self._username

    @username.setter
    def username(self, username: str):
        """Sets the username of this ConnectionSpec.

        Username  # noqa: E501

        :param username: The username of this ConnectionSpec.
        :type username: str
        """

        self._username = username

    @property
    def web_ui_link(self) -> str:
        """Gets the web_ui_link of this ConnectionSpec.

        Custom web UI link  # noqa: E501

        :return: The web_ui_link of this ConnectionSpec.
        :rtype: str
        """
        return self._web_ui_link

    @web_ui_link.setter
    def web_ui_link(self, web_ui_link: str):
        """Sets the web_ui_link of this ConnectionSpec.

        Custom web UI link  # noqa: E501

        :param web_ui_link: The web_ui_link of this ConnectionSpec.
        :type web_ui_link: str
        """

        self._web_ui_link = web_ui_link



#
#    Copyright 2020 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from typing import Dict, List

import docker
from docker.models.containers import Container

WORKSPACE_PATH = '/workspace'
TRAINING_DOCKER_LABELS: Dict[str, str] = {
    "component": "odahu",
    "api": "training"
}
PACKAGING_DOCKER_LABELS: Dict[str, str] = {
    "component": "odahu",
    "api": "packaging"
}


def raise_error_if_container_failed(container_id: str) -> None:
    client = docker.from_env()

    container_info = client.api.inspect_container(container_id)

    container_exit_code = container_info.get("State", {}).get("ExitCode", 0)
    if container_exit_code != 0:
        raise Exception(f'Container finished with {container_exit_code} error code')


def stream_container_logs(container: Container) -> None:
    """
    Stream logs of the Docker container to stdout
    :param container: Docker container
    """
    logs = container.logs(stream=True, follow=True)
    for log in logs:
        for line in log.splitlines():
            print(line.decode())


def convert_labels_to_filter(labels: Dict[str, str]) -> List[str]:
    """
    The docker client has the following label filter format: "key=value"
    :param labels: container labels
    :return: filters
    """
    return list(map(lambda item: f'{item[0]}={item[1]}', labels.items()))


def cleanup_docker_containers(labels: List[str]):
    client = docker.from_env()
    containers = client.containers.list(filters={
        "status": "exited",
        "label": labels,
    })

    if not containers:
        print('There are no containers')
        return

    for container in containers:
        print(f'{container.id} container has been deleted')
        container.remove()



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class ModelDeploymentStatus(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, available_replicas: int=None, created_at: str=None, deployment: str=None, last_revision_name: str=None, last_updated_time: str=None, replicas: int=None, service: str=None, service_url: str=None, state: str=None, updated_at: str=None):  # noqa: E501
        """ModelDeploymentStatus - a model defined in Swagger

        :param available_replicas: The available_replicas of this ModelDeploymentStatus.  # noqa: E501
        :type available_replicas: int
        :param created_at: The created_at of this ModelDeploymentStatus.  # noqa: E501
        :type created_at: str
        :param deployment: The deployment of this ModelDeploymentStatus.  # noqa: E501
        :type deployment: str
        :param last_revision_name: The last_revision_name of this ModelDeploymentStatus.  # noqa: E501
        :type last_revision_name: str
        :param last_updated_time: The last_updated_time of this ModelDeploymentStatus.  # noqa: E501
        :type last_updated_time: str
        :param replicas: The replicas of this ModelDeploymentStatus.  # noqa: E501
        :type replicas: int
        :param service: The service of this ModelDeploymentStatus.  # noqa: E501
        :type service: str
        :param service_url: The service_url of this ModelDeploymentStatus.  # noqa: E501
        :type service_url: str
        :param state: The state of this ModelDeploymentStatus.  # noqa: E501
        :type state: str
        :param updated_at: The updated_at of this ModelDeploymentStatus.  # noqa: E501
        :type updated_at: str
        """
        self.swagger_types = {
            'available_replicas': int,
            'created_at': str,
            'deployment': str,
            'last_revision_name': str,
            'last_updated_time': str,
            'replicas': int,
            'service': str,
            'service_url': str,
            'state': str,
            'updated_at': str
        }

        self.attribute_map = {
            'available_replicas': 'availableReplicas',
            'created_at': 'createdAt',
            'deployment': 'deployment',
            'last_revision_name': 'lastRevisionName',
            'last_updated_time': 'lastUpdatedTime',
            'replicas': 'replicas',
            'service': 'service',
            'service_url': 'serviceURL',
            'state': 'state',
            'updated_at': 'updatedAt'
        }

        self._available_replicas = available_replicas
        self._created_at = created_at
        self._deployment = deployment
        self._last_revision_name = last_revision_name
        self._last_updated_time = last_updated_time
        self._replicas = replicas
        self._service = service
        self._service_url = service_url
        self._state = state
        self._updated_at = updated_at

    @classmethod
    def from_dict(cls, dikt) -> 'ModelDeploymentStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ModelDeploymentStatus of this ModelDeploymentStatus.  # noqa: E501
        :rtype: ModelDeploymentStatus
        """
        return util.deserialize_model(dikt, cls)

    @property
    def available_replicas(self) -> int:
        """Gets the available_replicas of this ModelDeploymentStatus.

        Number of available pods  # noqa: E501

        :return: The available_replicas of this ModelDeploymentStatus.
        :rtype: int
        """
        return self._available_replicas

    @available_replicas.setter
    def available_replicas(self, available_replicas: int):
        """Sets the available_replicas of this ModelDeploymentStatus.

        Number of available pods  # noqa: E501

        :param available_replicas: The available_replicas of this ModelDeploymentStatus.
        :type available_replicas: int
        """

        self._available_replicas = available_replicas

    @property
    def created_at(self) -> str:
        """Gets the created_at of this ModelDeploymentStatus.


        :return: The created_at of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at: str):
        """Sets the created_at of this ModelDeploymentStatus.


        :param created_at: The created_at of this ModelDeploymentStatus.
        :type created_at: str
        """

        self._created_at = created_at

    @property
    def deployment(self) -> str:
        """Gets the deployment of this ModelDeploymentStatus.

        The model k8s deployment name  # noqa: E501

        :return: The deployment of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._deployment

    @deployment.setter
    def deployment(self, deployment: str):
        """Sets the deployment of this ModelDeploymentStatus.

        The model k8s deployment name  # noqa: E501

        :param deployment: The deployment of this ModelDeploymentStatus.
        :type deployment: str
        """

        self._deployment = deployment

    @property
    def last_revision_name(self) -> str:
        """Gets the last_revision_name of this ModelDeploymentStatus.

        Last applied ready knative revision  # noqa: E501

        :return: The last_revision_name of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._last_revision_name

    @last_revision_name.setter
    def last_revision_name(self, last_revision_name: str):
        """Sets the last_revision_name of this ModelDeploymentStatus.

        Last applied ready knative revision  # noqa: E501

        :param last_revision_name: The last_revision_name of this ModelDeploymentStatus.
        :type last_revision_name: str
        """

        self._last_revision_name = last_revision_name

    @property
    def last_updated_time(self) -> str:
        """Gets the last_updated_time of this ModelDeploymentStatus.

        Time when credentials was updated  # noqa: E501

        :return: The last_updated_time of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._last_updated_time

    @last_updated_time.setter
    def last_updated_time(self, last_updated_time: str):
        """Sets the last_updated_time of this ModelDeploymentStatus.

        Time when credentials was updated  # noqa: E501

        :param last_updated_time: The last_updated_time of this ModelDeploymentStatus.
        :type last_updated_time: str
        """

        self._last_updated_time = last_updated_time

    @property
    def replicas(self) -> int:
        """Gets the replicas of this ModelDeploymentStatus.

        Expected number of pods under current load  # noqa: E501

        :return: The replicas of this ModelDeploymentStatus.
        :rtype: int
        """
        return self._replicas

    @replicas.setter
    def replicas(self, replicas: int):
        """Sets the replicas of this ModelDeploymentStatus.

        Expected number of pods under current load  # noqa: E501

        :param replicas: The replicas of this ModelDeploymentStatus.
        :type replicas: int
        """

        self._replicas = replicas

    @property
    def service(self) -> str:
        """Gets the service of this ModelDeploymentStatus.

        The model k8s service name  # noqa: E501

        :return: The service of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._service

    @service.setter
    def service(self, service: str):
        """Sets the service of this ModelDeploymentStatus.

        The model k8s service name  # noqa: E501

        :param service: The service of this ModelDeploymentStatus.
        :type service: str
        """

        self._service = service

    @property
    def service_url(self) -> str:
        """Gets the service_url of this ModelDeploymentStatus.

        The model k8s service name  # noqa: E501

        :return: The service_url of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._service_url

    @service_url.setter
    def service_url(self, service_url: str):
        """Sets the service_url of this ModelDeploymentStatus.

        The model k8s service name  # noqa: E501

        :param service_url: The service_url of this ModelDeploymentStatus.
        :type service_url: str
        """

        self._service_url = service_url

    @property
    def state(self) -> str:
        """Gets the state of this ModelDeploymentStatus.

        The state of a model    \"Processing\" - A model was not deployed. Because some parameters of the                  custom resource are wrong. For example, there is not a model                  image in a Docker registry.   \"Ready\" - A model was deployed successfully.  # noqa: E501

        :return: The state of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state: str):
        """Sets the state of this ModelDeploymentStatus.

        The state of a model    \"Processing\" - A model was not deployed. Because some parameters of the                  custom resource are wrong. For example, there is not a model                  image in a Docker registry.   \"Ready\" - A model was deployed successfully.  # noqa: E501

        :param state: The state of this ModelDeploymentStatus.
        :type state: str
        """

        self._state = state

    @property
    def updated_at(self) -> str:
        """Gets the updated_at of this ModelDeploymentStatus.


        :return: The updated_at of this ModelDeploymentStatus.
        :rtype: str
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at: str):
        """Sets the updated_at of this ModelDeploymentStatus.


        :param updated_at: The updated_at of this ModelDeploymentStatus.
        :type updated_at: str
        """

        self._updated_at = updated_at



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models.json_schema import JsonSchema  # noqa: F401,E501
from odahuflow.sdk.models.target_schema import TargetSchema  # noqa: F401,E501
from odahuflow.sdk.models import util


class Schema(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, arguments: JsonSchema=None, targets: List[TargetSchema]=None):  # noqa: E501
        """Schema - a model defined in Swagger

        :param arguments: The arguments of this Schema.  # noqa: E501
        :type arguments: JsonSchema
        :param targets: The targets of this Schema.  # noqa: E501
        :type targets: List[TargetSchema]
        """
        self.swagger_types = {
            'arguments': JsonSchema,
            'targets': List[TargetSchema]
        }

        self.attribute_map = {
            'arguments': 'arguments',
            'targets': 'targets'
        }

        self._arguments = arguments
        self._targets = targets

    @classmethod
    def from_dict(cls, dikt) -> 'Schema':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Schema of this Schema.  # noqa: E501
        :rtype: Schema
        """
        return util.deserialize_model(dikt, cls)

    @property
    def arguments(self) -> JsonSchema:
        """Gets the arguments of this Schema.

        Arguments schema  # noqa: E501

        :return: The arguments of this Schema.
        :rtype: JsonSchema
        """
        return self._arguments

    @arguments.setter
    def arguments(self, arguments: JsonSchema):
        """Sets the arguments of this Schema.

        Arguments schema  # noqa: E501

        :param arguments: The arguments of this Schema.
        :type arguments: JsonSchema
        """

        self._arguments = arguments

    @property
    def targets(self) -> List[TargetSchema]:
        """Gets the targets of this Schema.

        Targets schema  # noqa: E501

        :return: The targets of this Schema.
        :rtype: List[TargetSchema]
        """
        return self._targets

    @targets.setter
    def targets(self, targets: List[TargetSchema]):
        """Sets the targets of this Schema.

        Targets schema  # noqa: E501

        :param targets: The targets of this Schema.
        :type targets: List[TargetSchema]
        """

        self._targets = targets



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from odahuflow.sdk.models.base_model_ import Model
from odahuflow.sdk.models import util


class ToolchainIntegrationSpec(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, additional_environments: object=None, default_image: str=None, entrypoint: str=None):  # noqa: E501
        """ToolchainIntegrationSpec - a model defined in Swagger

        :param additional_environments: The additional_environments of this ToolchainIntegrationSpec.  # noqa: E501
        :type additional_environments: object
        :param default_image: The default_image of this ToolchainIntegrationSpec.  # noqa: E501
        :type default_image: str
        :param entrypoint: The entrypoint of this ToolchainIntegrationSpec.  # noqa: E501
        :type entrypoint: str
        """
        self.swagger_types = {
            'additional_environments': object,
            'default_image': str,
            'entrypoint': str
        }

        self.attribute_map = {
            'additional_environments': 'additionalEnvironments',
            'default_image': 'defaultImage',
            'entrypoint': 'entrypoint'
        }

        self._additional_environments = additional_environments
        self._default_image = default_image
        self._entrypoint = entrypoint

    @classmethod
    def from_dict(cls, dikt) -> 'ToolchainIntegrationSpec':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ToolchainIntegrationSpec of this ToolchainIntegrationSpec.  # noqa: E501
        :rtype: ToolchainIntegrationSpec
        """
        return util.deserialize_model(dikt, cls)

    @property
    def additional_environments(self) -> object:
        """Gets the additional_environments of this ToolchainIntegrationSpec.

        Additional environments for a training process  # noqa: E501

        :return: The additional_environments of this ToolchainIntegrationSpec.
        :rtype: object
        """
        return self._additional_environments

    @additional_environments.setter
    def additional_environments(self, additional_environments: object):
        """Sets the additional_environments of this ToolchainIntegrationSpec.

        Additional environments for a training process  # noqa: E501

        :param additional_environments: The additional_environments of this ToolchainIntegrationSpec.
        :type additional_environments: object
        """

        self._additional_environments = additional_environments

    @property
    def default_image(self) -> str:
        """Gets the default_image of this ToolchainIntegrationSpec.

        Default training Docker image  # noqa: E501

        :return: The default_image of this ToolchainIntegrationSpec.
        :rtype: str
        """
        return self._default_image

    @default_image.setter
    def default_image(self, default_image: str):
        """Sets the default_image of this ToolchainIntegrationSpec.

        Default training Docker image  # noqa: E501

        :param default_image: The default_image of this ToolchainIntegrationSpec.
        :type default_image: str
        """

        self._default_image = default_image

    @property
    def entrypoint(self) -> str:
        """Gets the entrypoint of this ToolchainIntegrationSpec.

        Path to binary which starts a training process  # noqa: E501

        :return: The entrypoint of this ToolchainIntegrationSpec.
        :rtype: str
        """
        return self._entrypoint

    @entrypoint.setter
    def entrypoint(self, entrypoint: str):
        """Sets the entrypoint of this ToolchainIntegrationSpec.

        Path to binary which starts a training process  # noqa: E501

        :param entrypoint: The entrypoint of this ToolchainIntegrationSpec.
        :type entrypoint: str
        """

        self._entrypoint = entrypoint
